#define CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <CL/cl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

#ifndef LOCAL_SIZE_0
#define LOCAL_SIZE_0 8
#endif

#ifndef LOCAL_SIZE_1
#define LOCAL_SIZE_1 8
#endif

double main_clock;
double exec_clock;

#define CELL_AUTOM_KERNEL_NAME_MULT "Mult"

#define CELL_AUTOM_KERNEL_MULT \
    " __kernel void Mult(const int SIZE, \
            __global float *d_result, __global const float *d_matrix_a, __global const float *d_matrix_b ) \
    { \
		__local float tile_a[LOCAL_SIZE_0][LOCAL_SIZE_1]; \
		__local float tile_b[LOCAL_SIZE_0][LOCAL_SIZE_1]; \
        int row = get_global_id(1); \
		int col = get_global_id(0); \
		\
		int n_groups_x = get_num_groups(0); \
        \
        int local_size = get_local_size(0); \
        \
        int local_x = get_local_id(0); \
		int local_y = get_local_id(1); \
		\
        float tmp = 0.0; \
	    int idx; \
        \
        for (int sub = 0; sub < n_groups_x; ++sub) { \
		    idx = row * SIZE + sub * local_size + local_x; \
		    if (idx >= SIZE * SIZE) { \
			    tile_a[local_y][local_x] = 0; \
		    } else { \
			    tile_a[local_y][local_x] = d_matrix_a[idx]; \
		    } \
		    idx = (sub * local_size + local_y) * SIZE + col; \
		    if (idx >= SIZE * SIZE) { \
			    tile_b[local_y][local_x] = 0; \
		    } else { \
			    tile_b[local_y][local_x] = d_matrix_b[idx]; \
		    } \
		    barrier(CLK_LOCAL_MEM_FENCE); \
            \
		    for (int k = 0; k < local_size; ++k) { \
                tmp += tile_a[local_y][k] * tile_b[k][local_x]; \
            } \
		    barrier(CLK_LOCAL_MEM_FENCE); \
	    } \
        if (row < SIZE && col < SIZE) { \
            d_result[row * SIZE + col] = tmp; \
        } \
    } "


int main (int argc, char *argv[]) {
	main_clock = omp_get_wtime();

	if (argc < 5) {
		fprintf(stderr, "Usage: %s <size> <numIter> <platform> <device>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	cl_int SIZE = atoi(argv[1]);
	cl_int N_ITER = atoi(argv[2]);
	cl_int PLATFORM = atoi(argv[3]);
	cl_int DEVICE = atoi(argv[4]);

	size_t MATRIX_SIZE = sizeof(cl_float) * SIZE * SIZE;
	size_t MATRIX_TAM = SIZE * SIZE;

	cl_int err;

    char *kernel_raw_mult = CELL_AUTOM_KERNEL_MULT;
	
	char *kernel_raw = (char *)malloc((strlen(kernel_raw_mult) + 300) * sizeof(char));
	sprintf(&kernel_raw[0], " #define LOCAL_SIZE_0  %d \n#define LOCAL_SIZE_1 %d \n", LOCAL_SIZE_0, LOCAL_SIZE_1);
	strcat(kernel_raw, kernel_raw_mult);

	size_t kernel_size = strlen(kernel_raw);

	size_t local_size[2];
	size_t global_size[2];

	double result;
	double sum;

	cl_platform_id platform_id;
	cl_device_id device_id;

	cl_context context;
	cl_program program;
	cl_kernel kernel_mult;
	cl_kernel kernel_copy;

	cl_command_queue_properties properties;
	
	cl_command_queue main_command_queue;
	cl_command_queue write_command_queue;
	cl_command_queue read_command_queue;

	cl_mem mem_matrix_a;
    cl_mem mem_matrix_b;
	cl_mem mem_matrix_c;

	cl_mem pinned_matrix_a;
    cl_mem pinned_matrix_b;
	cl_mem pinned_matrix_c;

	cl_float *p_pinned_matrix_a;
    cl_float *p_pinned_matrix_b;
	cl_float *p_pinned_matrix_c;

	cl_event event_read_even;
	cl_event event_read_odd;
	cl_event event_write;
	cl_event event_mult;
    cl_event event_base;

	cl_event wait_list[5];

    cl_platform_id *p_platforms = (cl_platform_id *)malloc((PLATFORM + 1) * sizeof(cl_platform_id));
	clGetPlatformIDs(PLATFORM + 1, p_platforms, NULL);
    platform_id = p_platforms[PLATFORM];
	free(p_platforms);

    cl_device_id *p_devices = (cl_device_id *)malloc((DEVICE + 1) * sizeof(cl_device_id));
    clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, DEVICE + 1, p_devices, NULL);
    device_id = p_devices[DEVICE];
    free(p_devices);

	cl_context_properties context_properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_id, 0 };
	context = clCreateContext(context_properties, 1, &device_id, NULL, NULL, &err);

	main_command_queue = clCreateCommandQueue(context, device_id, 0, &err);
	read_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
	write_command_queue = clCreateCommandQueue(context, device_id, properties, &err);

	pinned_matrix_a = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	p_pinned_matrix_a = (float *)clEnqueueMapBuffer(main_command_queue, pinned_matrix_a, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	mem_matrix_a = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);

    pinned_matrix_b = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	p_pinned_matrix_b = (float *)clEnqueueMapBuffer(main_command_queue, pinned_matrix_b, CL_TRUE, CL_MAP_WRITE | CL_MAP_READ, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	mem_matrix_b = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);

	pinned_matrix_c = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	p_pinned_matrix_c = (float *)clEnqueueMapBuffer(main_command_queue, pinned_matrix_c, CL_TRUE, CL_MAP_WRITE | CL_MAP_READ, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	mem_matrix_c = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);

	program = clCreateProgramWithSource(context, 1, (const char **)(&kernel_raw), (const size_t *)(&kernel_size), &err);
	clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
	kernel_mult = clCreateKernel(program, CELL_AUTOM_KERNEL_NAME_MULT, &err);

	local_size[0] = LOCAL_SIZE_0;
	local_size[1] = LOCAL_SIZE_1;

	global_size[0] = global_size[1] = SIZE;

	if ((SIZE % local_size[0]) != 0) {
		global_size[0] += (local_size[0] - (SIZE % local_size[0]));
	}

	if ((SIZE % local_size[1]) != 0) {
		global_size[1] += (local_size[1] - (SIZE % local_size[1]));
	}

	event_base = clCreateUserEvent(context, &err);
	clSetUserEventStatus(event_base, CL_COMPLETE);

	event_read_even = event_base;
	event_read_odd = event_base;
    event_mult = event_base;
	event_write = event_base;

	clRetainEvent(event_read_even);
	clRetainEvent(event_read_odd);
	clRetainEvent(event_mult);
	clRetainEvent(event_write);

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			p_pinned_matrix_a[i * SIZE + j] = 0.001;
			p_pinned_matrix_b[i * SIZE + j] = 0.001;
			p_pinned_matrix_c[i * SIZE + j] = 0;
		}
	}

	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);
	

	clReleaseEvent(event_write);
	clEnqueueWriteBuffer(write_command_queue, mem_matrix_a, CL_FALSE, 0, MATRIX_SIZE,(void *)p_pinned_matrix_a, 0, NULL, &event_write);
	
	clReleaseEvent(event_write);
	clEnqueueWriteBuffer(write_command_queue, mem_matrix_b, CL_FALSE, 0, MATRIX_SIZE,(void *)p_pinned_matrix_b, 0, NULL, &event_write);

	clReleaseEvent(event_write);
	clEnqueueWriteBuffer(write_command_queue, mem_matrix_c, CL_FALSE, 0, MATRIX_SIZE,(void *)p_pinned_matrix_c, 0, NULL, &event_write);
		
	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);
	
	exec_clock = omp_get_wtime();

	for (int i = 0; i < N_ITER; i++) {
		if ((i % 2) == 0) {
			clSetKernelArg(kernel_mult, 0, sizeof(cl_int), &SIZE);
			clSetKernelArg(kernel_mult, 1, sizeof(cl_mem), &mem_matrix_c);
			clSetKernelArg(kernel_mult, 2, sizeof(cl_mem), &mem_matrix_a);
			clSetKernelArg(kernel_mult, 3, sizeof(cl_mem), &mem_matrix_b);

			clReleaseEvent(event_mult);
			clEnqueueNDRangeKernel(main_command_queue, kernel_mult, 2, NULL, global_size, local_size, 1, &event_read_odd, &event_mult);
			clFlush(main_command_queue);

			clReleaseEvent(event_read_odd);
			clEnqueueReadBuffer(read_command_queue, mem_matrix_c, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_matrix_c, 1, &event_mult, &event_read_odd);
			clFlush(read_command_queue);
		} else {
			clSetKernelArg(kernel_mult, 0, sizeof(cl_int), &SIZE);
			clSetKernelArg(kernel_mult, 1, sizeof(cl_mem), &mem_matrix_b);
			clSetKernelArg(kernel_mult, 2, sizeof(cl_mem), &mem_matrix_a);
			clSetKernelArg(kernel_mult, 3, sizeof(cl_mem), &mem_matrix_c);

			clReleaseEvent(event_mult);
			clEnqueueNDRangeKernel(main_command_queue, kernel_mult, 2, NULL, global_size, local_size, 1, &event_read_even, &event_mult);
			clFlush(main_command_queue);

			clReleaseEvent(event_read_even);
			clEnqueueReadBuffer(read_command_queue, mem_matrix_b, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_matrix_b, 1, &event_mult, &event_read_even);
			clFlush(read_command_queue);
		}
	}

	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);
	
	exec_clock = omp_get_wtime() - exec_clock ;

	sum = result = 0;
	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			if ((N_ITER % 2) == 0) {
				sum += pow(p_pinned_matrix_b[i * SIZE + j], 2);
			} else {
				sum += pow(p_pinned_matrix_c[i * SIZE + j], 2);
			}
		}
	}
	result = sqrt(sum);
	
	printf("\n ---------------------- RESULT ---------------------- \n");
	printf("\n Sum: %lf", sum);
	printf("\n Result: %lf", result);
	printf("\n\n ---------------------------------------------------- \n");
	
	clReleaseEvent(event_mult);
	clReleaseEvent(event_read_even);
	clReleaseEvent(event_read_odd);
	clReleaseEvent(event_write);
	clReleaseEvent(event_base);

	clEnqueueUnmapMemObject(main_command_queue, pinned_matrix_a, p_pinned_matrix_a, 0, NULL, NULL);
	clEnqueueUnmapMemObject(main_command_queue, pinned_matrix_b, p_pinned_matrix_b, 0, NULL, NULL);
	clEnqueueUnmapMemObject(main_command_queue, pinned_matrix_c, p_pinned_matrix_c, 0, NULL, NULL);
	
	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clReleaseMemObject(pinned_matrix_a);
	clReleaseMemObject(pinned_matrix_b);
	clReleaseMemObject(pinned_matrix_c);
	clReleaseMemObject(mem_matrix_a);
	clReleaseMemObject(mem_matrix_b);
	clReleaseMemObject(mem_matrix_c);

	clReleaseKernel(kernel_mult);
	clReleaseProgram(program);
	
	clReleaseCommandQueue(main_command_queue);
	clReleaseCommandQueue(read_command_queue);
	clReleaseCommandQueue(write_command_queue);
	
	clReleaseContext(context);

	main_clock = omp_get_wtime() - main_clock ;
	
	printf("\n ---------------------- TIMERS ---------------------- \n");
	printf("Clock main: %lf\n", main_clock);
	printf("Clock exec: %lf\n", exec_clock);
	printf("\n\n ---------------------------------------------------- \n");
	
	return 0;
}

