#define CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <CL/cl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h> //For exit()
#include <math.h>
#include <omp.h>

#define SEED 6834723

#ifndef LOCAL_SIZE_0
#define LOCAL_SIZE_0 8
#endif

#ifndef LOCAL_SIZE_1
#define LOCAL_SIZE_1 8
#endif

#ifdef _CTRL_EXAMPLES_OPENCL_GPU_ERROR_CHECK_
	#include <assert.h>
	
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_DEBUG_
		#define OPENCL_ASSERT_OP( operation ) \
			err = operation; \
			printf("error: %d\n", err); \
			assert(err == CL_SUCCESS);

		#define OPENCL_ASSERT_ERROR( err ) \
			printf("error: %d\n", err); \
			assert(err == CL_SUCCESS)

	#else 
		#define OPENCL_ASSERT_OP( operation ) \
			assert(operation == CL_SUCCESS);

		#define OPENCL_ASSERT_ERROR( err ) \
			assert(err == CL_SUCCESS)
	#endif
#else
	#define OPENCL_ASSERT_OP( operation ) operation
	#define OPENCL_ASSERT_ERROR(err)
#endif


#define OPENCL_PROFILE_OP_READ 1
#define OPENCL_PROFILE_OP_WRITE 2
#define OPENCL_PROFILE_OP_MULT 3
#define OPENCL_PROFILE_OP_COPY 4

#define OPENCL_PROFILE_QUEUE_MAIN 0
#define OPENCL_PROFILE_QUEUE_READ 1
#define OPENCL_PROFILE_QUEUE_WRITE 2

#ifdef _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
	#define OPENCL_PROFILE_READ( event )	\
		profiling_read_events[i_read_task] = event; \
		OPENCL_ASSERT_OP( clRetainEvent(profiling_read_events[i_read_task]) ); \
		i_read_task++;

	#define OPENCL_PROFILE_WRITE( event )	\
		profiling_write_events[i_write_task] = event; \
		OPENCL_ASSERT_OP( clRetainEvent(profiling_write_events[i_write_task]) ); \
		i_write_task++;

	#define OPENCL_PROFILE_MULT( event ) \
		profiling_mult_events[i_mult_task] = event; \
		OPENCL_ASSERT_OP( clRetainEvent(profiling_mult_events[i_mult_task]) ); \
		i_mult_task++;

	#define OPENCL_PROFILE_COPY( event ) \
		profiling_copy_events[i_copy_task] = event; \
		OPENCL_ASSERT_OP( clRetainEvent(profiling_copy_events[i_copy_task]) ); \
		i_copy_task++;

	#define OPENCL_PROFILE_FIRST( event ) \
		first_profiling_event = event; \
		OPENCL_ASSERT_OP( clRetainEvent(first_profiling_event) );

	#define OPENCL_PROFILE_LAST( event ) \
		last_profiling_event = event; \
		OPENCL_ASSERT_OP( clRetainEvent(last_profiling_event) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_GENERATE_VISUAL_PROFILER_
		typedef struct {
			cl_event event;
			int queue;
			int	op;
		} visual_event;

		#define OPENCL_PROFILE_VISUAL( command_queue, visual_event, operation ) \
			profiling_visual_events[i_visual_task].event = visual_event; \
			profiling_visual_events[i_visual_task].queue = command_queue; \
			profiling_visual_events[i_visual_task].op = operation; \
			OPENCL_ASSERT_OP( clRetainEvent(profiling_visual_events[i_visual_task].event) ); \
			i_visual_task++;	
	#else 
		#define OPENCL_PROFILE_VISUAL( queue, event, op ) 
	#endif
#else
	#define OPENCL_PROFILE_READ( event )
	#define OPENCL_PROFILE_WRITE( event )
	#define OPENCL_PROFILE_MULT( event )
	#define OPENCL_PROFILE_COPY( event )
	#define OPENCL_PROFILE_FIRST( event )
	#define OPENCL_PROFILE_LAST( event )
	#define OPENCL_PROFILE_VISUAL( queue, event, op )
#endif

double main_clock;
double exec_clock;


#define CELL_AUTOM_KERNEL_NAME_MULT "Mult"

#define CELL_AUTOM_KERNEL_MULT \
    " __kernel void Mult(const int SIZE, \
            __global float *d_result, __global const float *d_matrix_a, __global const float *d_matrix_b ) \
    { \
		__local float tile_a[LOCAL_SIZE_0][LOCAL_SIZE_1]; \
		__local float tile_b[LOCAL_SIZE_0][LOCAL_SIZE_1]; \
        int row = get_global_id(1); \
		int col = get_global_id(0); \
		\
		int n_groups_x = get_num_groups(0); \
        \
        int local_size = get_local_size(0); \
        \
        int local_x = get_local_id(0); \
		int local_y = get_local_id(1); \
		\
        float tmp = 0.0; \
	    int idx; \
        \
        for (int sub = 0; sub < n_groups_x; ++sub) { \
		    idx = row * SIZE + sub * local_size + local_x; \
		    if (idx >= SIZE * SIZE) { \
			    tile_a[local_y][local_x] = 0; \
		    } else { \
			    tile_a[local_y][local_x] = d_matrix_a[idx]; \
		    } \
		    idx = (sub * local_size + local_y) * SIZE + col; \
		    if (idx >= SIZE * SIZE) { \
			    tile_b[local_y][local_x] = 0; \
		    } else { \
			    tile_b[local_y][local_x] = d_matrix_b[idx]; \
		    } \
		    barrier(CLK_LOCAL_MEM_FENCE); \
            \
		    for (int k = 0; k < local_size; ++k) { \
                tmp += tile_a[local_y][k] * tile_b[k][local_x]; \
            } \
		    barrier(CLK_LOCAL_MEM_FENCE); \
	    } \
        if (row < SIZE && col < SIZE) { \
            d_result[row * SIZE + col] = tmp; \
        } \
    } "


#define CELL_AUTOM_KERNEL_NAME_COPY "Copy"

#define CELL_AUTOM_KERNEL_COPY \
    " __kernel void Copy(const int SIZE, __global float *d_copy, __global const float *d_result) { \
		int row = get_global_id(1); \
		int col = get_global_id(0); \
        \
		if (row < SIZE && col < SIZE) { \
            d_copy[row * SIZE + col] = d_result[row * SIZE + col]; \
        } \
    } "


int main (int argc, char *argv[]) {
	main_clock = omp_get_wtime();

	/* ARGUMENTS */

	if (argc < 5) {
		fprintf(stderr, "Usage: %s <size> <numIter> <platform> <device>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	cl_int SIZE = atoi(argv[1]);
	cl_int N_ITER = atoi(argv[2]);
	cl_int PLATFORM = atoi(argv[3]);
	cl_int DEVICE = atoi(argv[4]);

	size_t MATRIX_SIZE = sizeof(cl_float) * SIZE * SIZE;
	size_t MATRIX_TAM = SIZE * SIZE;

	cl_int n_read_taks = N_ITER;
	cl_int n_write_taks = 2;
	cl_int n_mult_taks = N_ITER;
	cl_int n_copy_taks = N_ITER;

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
			printf("%ld %ld %d %d\n", SIZE, N_ITER, PLATFORM, DEVICE);
		#else
			printf("\n ----------------------- ARGS ----------------------- \n");
			printf("\n SIZE: %d", SIZE);
			printf("\n N_ITER: %d", N_ITER);
			printf("\n PLATFORM: %d", PLATFORM);
			printf("\n DEVICE: %d", DEVICE);
			printf("\n TOTAL TASKS: %d", n_mult_taks + n_copy_taks + n_read_taks + n_write_taks);
			printf("\n KERNEL TASKS: %d", n_mult_taks + n_copy_taks);
			printf("\n MULT TASKS: %d", n_mult_taks);
			printf("\n COPY TASKS: %d", n_copy_taks);
			printf("\n OFFLOADING TASKS: %d", n_read_taks + n_write_taks);
			printf("\n READ TASKS: %d", n_read_taks);
			printf("\n WRITE TASKS: %d", n_write_taks);
			#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
				printf("\n POLICY ASYNC");
			#else // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
				printf("\n POLICY SYNC");
			#endif // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			printf("\n\n ---------------------------------------------------- \n");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

	/* VARIABLES */

	cl_int err;

    char *kernel_raw_mult = CELL_AUTOM_KERNEL_MULT;
	char *kernel_raw_copy = CELL_AUTOM_KERNEL_COPY;

	char *kernel_raw = (char *)malloc((strlen(kernel_raw_mult) + strlen(kernel_raw_copy) + 300) * sizeof(char));
	sprintf(&kernel_raw[0], " #define LOCAL_SIZE_0  %d \n#define LOCAL_SIZE_1 %d \n", LOCAL_SIZE_0, LOCAL_SIZE_1);
	strcat(kernel_raw, kernel_raw_copy);
	strcat(kernel_raw, kernel_raw_mult);

	size_t kernel_size = strlen(kernel_raw);

	size_t local_size[2];
	size_t global_size[2];

	double result;
	double sum;

	cl_platform_id platform_id;
	cl_device_id device_id;

	cl_context context;
	cl_program program;
	cl_kernel kernel_mult;
	cl_kernel kernel_copy;

	cl_command_queue_properties properties;
	
	cl_command_queue main_command_queue;
	cl_command_queue write_command_queue;
	cl_command_queue read_command_queue;

	cl_mem mem_matrix_a;
    cl_mem mem_matrix_b;
	cl_mem mem_matrix_copy;

	cl_mem pinned_matrix_a;
    cl_mem pinned_matrix_b;

	cl_float *p_pinned_matrix_a;
    cl_float *p_pinned_matrix_b;

	cl_event event_matrix_a_write;
    cl_event event_matrix_a_read;

	cl_event event_matrix_b_write;

	cl_event event_copy;
	cl_event event_mult;

    cl_event event_base;

	cl_event wait_list[5];

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
		cl_profiling_info profiling_info_start = CL_PROFILING_COMMAND_START;
		cl_profiling_info profiling_info_end = CL_PROFILING_COMMAND_END;

		cl_event first_profiling_event;
		cl_event last_profiling_event;

		cl_event *profiling_read_events = (cl_event *)malloc(n_read_taks * sizeof(cl_event));
		cl_event *profiling_write_events = (cl_event *)malloc(n_write_taks * sizeof(cl_event));
		cl_event *profiling_mult_events = (cl_event *)malloc(n_mult_taks * sizeof(cl_event));
		cl_event *profiling_copy_events = (cl_event *)malloc(n_copy_taks * sizeof(cl_event));

		int i_read_task = 0;
		int i_write_task = 0;
		int i_mult_task = 0;
		int i_copy_task = 0;


		cl_ulong profiling_total = 0;
		cl_ulong profiling_sum = 0;
		cl_ulong profiling_offloading = 0;
		cl_ulong profiling_read = 0;
		cl_ulong profiling_write = 0;
		cl_ulong profiling_kernel = 0;
		cl_ulong profiling_mult = 0;
		cl_ulong profiling_copy = 0;

		cl_ulong profiling_start = 0;
		cl_ulong profiling_end = 0;
		
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_GENERATE_VISUAL_PROFILER_
			int n_visual_task = n_mult_taks + n_copy_taks + n_read_taks + n_write_taks;
			int i_visual_task = 0;

			visual_event *profiling_visual_events = (visual_event *)malloc(n_visual_task * sizeof(visual_event));
		#endif
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_

	/* PLATFORMS & DEVICES */

    cl_platform_id *p_platforms = (cl_platform_id *)malloc((PLATFORM + 1) * sizeof(cl_platform_id));
	OPENCL_ASSERT_OP( clGetPlatformIDs(PLATFORM + 1, p_platforms, NULL) );
    platform_id = p_platforms[PLATFORM];
	free(p_platforms);

    cl_device_id *p_devices = (cl_device_id *)malloc((DEVICE + 1) * sizeof(cl_device_id));
    OPENCL_ASSERT_OP( clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, DEVICE + 1, p_devices, NULL) );
    device_id = p_devices[DEVICE];
    free(p_devices);
	
	/* SET UP, CONTEXTO, COLAS, KERNELS, ETC */

	cl_context_properties context_properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_id, 0 };
	context = clCreateContext(context_properties, 1, &device_id, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	properties = 0;

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
		properties |= CL_QUEUE_PROFILING_ENABLE;
	#endif

	main_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
	OPENCL_ASSERT_ERROR( err );
	
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		read_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
		OPENCL_ASSERT_ERROR( err );
		write_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
		OPENCL_ASSERT_ERROR( err );
	#endif

	pinned_matrix_a = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	p_pinned_matrix_a = (float *)clEnqueueMapBuffer(main_command_queue, pinned_matrix_a, CL_FALSE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	mem_matrix_a = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

    pinned_matrix_b = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR | CL_MEM_HOST_WRITE_ONLY, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	p_pinned_matrix_b = (float *)clEnqueueMapBuffer(main_command_queue, pinned_matrix_b, CL_FALSE, CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	mem_matrix_b = clCreateBuffer(context, CL_MEM_READ_ONLY| CL_MEM_HOST_WRITE_ONLY, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	mem_matrix_copy = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_HOST_NO_ACCESS, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	program = clCreateProgramWithSource(context, 1, (const char **)(&kernel_raw), (const size_t *)(&kernel_size), &err);
	OPENCL_ASSERT_ERROR( err );
	
	
	// OPENCL_ASSERT_OP( clBuildProgram(program, 1, &device_id, NULL, NULL, NULL) );
	err = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);

	if (err == CL_BUILD_PROGRAM_FAILURE) {
		size_t log_size;
		clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
		char *log = (char *) malloc(log_size);
		clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, log_size, log, NULL);
		printf("%s\n", log);
		free(log);
	}

	kernel_mult = clCreateKernel(program, CELL_AUTOM_KERNEL_NAME_MULT, &err);
	OPENCL_ASSERT_ERROR( err );
	
	kernel_copy = clCreateKernel(program, CELL_AUTOM_KERNEL_NAME_COPY, &err);
	OPENCL_ASSERT_ERROR( err );

	local_size[0] = LOCAL_SIZE_0;
	local_size[1] = LOCAL_SIZE_1;

	global_size[0] = global_size[1] = SIZE;

	if ((SIZE % local_size[0]) != 0) {
		global_size[0] += (local_size[0] - (SIZE % local_size[0]));
	}

	if ((SIZE % local_size[1]) != 0) {
		global_size[1] += (local_size[1] - (SIZE % local_size[1]));
	}

	err = clSetKernelArg(kernel_copy, 0, sizeof(cl_int), &SIZE);
	err |= clSetKernelArg(kernel_copy, 1, sizeof(cl_mem), &mem_matrix_a);
	err |= clSetKernelArg(kernel_copy, 2, sizeof(cl_mem), &mem_matrix_copy);
	OPENCL_ASSERT_ERROR( err );

	err = clSetKernelArg(kernel_mult, 0, sizeof(cl_int), &SIZE);
	err |= clSetKernelArg(kernel_mult, 1, sizeof(cl_mem), &mem_matrix_copy);
	err |= clSetKernelArg(kernel_mult, 2, sizeof(cl_mem), &mem_matrix_a);
    err |= clSetKernelArg(kernel_mult, 3, sizeof(cl_mem), &mem_matrix_b);
	OPENCL_ASSERT_ERROR( err );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
			printf("%lu %lu %lu %lu\n", local_size[0], local_size[1], global_size[0], global_size[1]);
		#else
			printf("\n ------------------ WORK-GROUP-INFO ----------------- \n");
			printf("\n Local Size: %lu x %lu", local_size[0], local_size[1]);
			printf("\n Global Size: %lu x %lu", global_size[0], global_size[1]);
			printf("\n\n ---------------------------------------------------- \n");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

	
	event_base = clCreateUserEvent(context, &err);
	OPENCL_ASSERT_ERROR( err );

	OPENCL_ASSERT_OP( clSetUserEventStatus(event_base, CL_COMPLETE) );

	event_copy = event_base;
    event_mult = event_base;
	event_matrix_a_read = event_base;

	OPENCL_ASSERT_OP( clRetainEvent(event_copy) );
	OPENCL_ASSERT_OP( clRetainEvent(event_mult) );
	OPENCL_ASSERT_OP( clRetainEvent(event_matrix_a_read) );

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			p_pinned_matrix_a[i * SIZE + j] = 2.0;
			p_pinned_matrix_b[i * SIZE + j] = 2.0;
		}
	}

	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(write_command_queue, mem_matrix_a, CL_FALSE, 0, MATRIX_SIZE,(void *)p_pinned_matrix_a, 0, NULL, &event_matrix_a_write) );
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_WRITE, event_matrix_a_write, OPENCL_PROFILE_OP_WRITE );
	#else
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(main_command_queue, mem_matrix_a, CL_FALSE, 0, MATRIX_SIZE,(void *)p_pinned_matrix_a, 0, NULL, &event_matrix_a_write) );
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_matrix_a_write, OPENCL_PROFILE_OP_WRITE );
	#endif
	OPENCL_PROFILE_WRITE( event_matrix_a_write );
	OPENCL_PROFILE_FIRST( event_matrix_a_write );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(write_command_queue, mem_matrix_b, CL_FALSE, 0, MATRIX_SIZE,(void *)p_pinned_matrix_b, 0, NULL, &event_matrix_b_write) );
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_WRITE, event_matrix_b_write, OPENCL_PROFILE_OP_WRITE );
	#else
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(main_command_queue, mem_matrix_b, CL_FALSE, 0, MATRIX_SIZE,(void *)p_pinned_matrix_b, 0, NULL, &event_matrix_b_write) );
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_matrix_b_write, OPENCL_PROFILE_OP_WRITE );
	#endif
	OPENCL_PROFILE_WRITE( event_matrix_b_write );
	
	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif

	exec_clock = omp_get_wtime();

	for (int i = 0; i < N_ITER; i++) {
			
		OPENCL_ASSERT_OP( clReleaseEvent(event_mult) );
		if (i != 0) {
			OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_mult, 2, NULL, global_size, local_size, 1, &event_copy, &event_mult) );
		} else {
			wait_list[0] = event_matrix_a_write;
			wait_list[1] = event_matrix_b_write;
			OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_mult, 2, NULL, global_size, local_size, 2, wait_list, &event_mult) );
		}
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_mult, OPENCL_PROFILE_OP_MULT );
		OPENCL_PROFILE_MULT( event_mult );

		OPENCL_ASSERT_OP( clFlush(main_command_queue) );

		wait_list[0] = event_mult;
		wait_list[1] = event_matrix_a_read;

		OPENCL_ASSERT_OP( clReleaseEvent(event_copy) );
		OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_copy, 2, NULL, global_size, local_size, 2, wait_list, &event_copy) );
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_copy, OPENCL_PROFILE_OP_COPY );
		OPENCL_PROFILE_COPY( event_copy );

		OPENCL_ASSERT_OP( clFlush(main_command_queue) );

		OPENCL_ASSERT_OP( clReleaseEvent(event_matrix_a_read) );
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			OPENCL_ASSERT_OP( clEnqueueReadBuffer(read_command_queue, mem_matrix_a, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_matrix_a, 1, &event_copy, &event_matrix_a_read) );
			OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_READ, event_matrix_a_read, OPENCL_PROFILE_OP_READ );
		#else
			OPENCL_ASSERT_OP( clEnqueueReadBuffer(main_command_queue, mem_matrix_a, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_matrix_a, 1, &event_copy, &event_matrix_a_read) );
			OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_matrix_a_read, OPENCL_PROFILE_OP_READ );
		#endif
		OPENCL_PROFILE_READ( event_matrix_a_read );

		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		#else
			OPENCL_ASSERT_OP( clFlush(main_command_queue) );
		#endif
	}

	OPENCL_PROFILE_LAST( event_matrix_a_read );

	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif

	exec_clock = omp_get_wtime() - exec_clock ;

	/* CALCULO RESULTADOS */

	sum = result = 0;
	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			sum += pow(p_pinned_matrix_a[i * SIZE + j], 2);
		}
	}
	result = sqrt(sum);

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
		printf("%lf %lf\n", sum, result);
	#else 
		printf("\n ---------------------- RESULT ---------------------- \n");
		printf("\n Sum: %lf", sum);
		printf("\n Result: %lf", result);
		printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_

	
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_GENERATE_VISUAL_PROFILER_
			FILE *f = fopen("visual_profiler_info.txt", "w");

			for (int i = 0; i < n_visual_task; i++) {
				OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_visual_events[i].event, profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
				OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_visual_events[i].event, profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );

				fprintf(f, "%d; %d; %lu; %lu\n", profiling_visual_events[i].queue, profiling_visual_events[i].op, profiling_start, profiling_end);
				OPENCL_ASSERT_OP( clReleaseEvent(profiling_visual_events[i].event) );
			}

			fclose(f);
			free(profiling_visual_events);
		#endif
		for (int i = 0; i < n_read_taks; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_read_events[i], profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_read_events[i], profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );

			profiling_sum += profiling_end - profiling_start;
			profiling_read += profiling_end - profiling_start;
			profiling_offloading += profiling_end - profiling_start;
			
			OPENCL_ASSERT_OP( clReleaseEvent(profiling_read_events[i]) );
		}
		free(profiling_read_events);

		for (int i = 0; i < n_write_taks; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_write_events[i], profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_write_events[i], profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );

			profiling_sum += profiling_end - profiling_start;
			profiling_write += profiling_end - profiling_start;
			profiling_offloading += profiling_end - profiling_start;
			
			OPENCL_ASSERT_OP( clReleaseEvent(profiling_write_events[i]) );
		}
		free(profiling_write_events);

		for (int i = 0; i < n_mult_taks; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_mult_events[i], profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_mult_events[i], profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );
			
			profiling_sum += profiling_end - profiling_start;
			profiling_mult += profiling_end - profiling_start;
			profiling_kernel += profiling_end - profiling_start;

			OPENCL_ASSERT_OP( clReleaseEvent(profiling_mult_events[i]) );
		}
		free(profiling_mult_events);

		for (int i = 0; i < n_copy_taks; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_copy_events[i], profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_copy_events[i], profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );
			
			profiling_sum += profiling_end - profiling_start;
			profiling_copy += profiling_end - profiling_start;
			profiling_kernel += profiling_end - profiling_start;

			OPENCL_ASSERT_OP( clReleaseEvent(profiling_copy_events[i]) );
		}
		free(profiling_copy_events);

		OPENCL_ASSERT_OP( clGetEventProfilingInfo(first_profiling_event, profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
		OPENCL_ASSERT_OP( clGetEventProfilingInfo(last_profiling_event, profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );

		profiling_total = profiling_end - profiling_start;

		OPENCL_ASSERT_OP( clReleaseEvent(first_profiling_event) );
		OPENCL_ASSERT_OP( clReleaseEvent(last_profiling_event) );

		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
			printf("%.3lf %.3lf %.3lf %.3lf %.3lf\n", profiling_total, profiling_sum, profiling_kernel, profiling_offloading, profiling_read, profiling_write);
		#else
			printf("\n -------------------- PROFILING --------------------- \n");
			printf("\n Total Time: %.3lf", profiling_total / 1000000.0);
			printf("\n Acumulate Sum Time: %.3lf", profiling_sum / 1000000.0);
			printf("\n Kernel Time: %.3lf", profiling_kernel / 1000000.0);
			printf("\n Mult Time: %.3lf", profiling_mult / 1000000.0);
			printf("\n Copy Time: %.3lf", profiling_copy / 1000000.0);
			printf("\n Offloading Time: %.3lf", profiling_offloading / 1000000.0);
			printf("\n Read Time: %.3lf", profiling_read / 1000000.0);
			printf("\n Write Time: %.3lf", profiling_write / 1000000.0);
			printf("\n\n ---------------------------------------------------- \n");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_

	/* RELEASE ZONE */
	
	OPENCL_ASSERT_OP( clReleaseEvent(event_copy) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_mult) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_matrix_a_read) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_matrix_a_write) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_matrix_b_write) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_base) );

	OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(main_command_queue, pinned_matrix_a, p_pinned_matrix_a, 0, NULL, NULL) );
	OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(main_command_queue, pinned_matrix_b, p_pinned_matrix_b, 0, NULL, NULL) );
	
	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	OPENCL_ASSERT_OP( clReleaseMemObject(pinned_matrix_a) );
	OPENCL_ASSERT_OP( clReleaseMemObject(pinned_matrix_b) );
	OPENCL_ASSERT_OP( clReleaseMemObject(mem_matrix_a) );
	OPENCL_ASSERT_OP( clReleaseMemObject(mem_matrix_b) );
	OPENCL_ASSERT_OP( clReleaseMemObject(mem_matrix_copy) );

	OPENCL_ASSERT_OP( clReleaseKernel(kernel_mult) );
	OPENCL_ASSERT_OP( clReleaseKernel(kernel_copy) );
	OPENCL_ASSERT_OP( clReleaseProgram(program) );
	
	OPENCL_ASSERT_OP( clReleaseCommandQueue(main_command_queue) );
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clReleaseCommandQueue(read_command_queue) );
		OPENCL_ASSERT_OP( clReleaseCommandQueue(write_command_queue) );
	#endif
	
	OPENCL_ASSERT_OP( clReleaseContext(context) );

	main_clock = omp_get_wtime() - main_clock ;

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
		printf("%lf %lf\n", main_clock, exec_clock);
	#else 
		printf("\n ---------------------- TIMERS ---------------------- \n");
		printf("Clock main: %lf\n", main_clock);
		printf("Clock exec: %lf\n", exec_clock);
		printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	
	return 0;
}

