#define CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <CL/cl.h>
#include <math.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define IMAD(a, b, c) (mul24((a), (b)) + (c))
#define cKernelRadius 1
#define rKernelRadius 1
#define alpha 1.0

#ifndef LOCAL_SIZE_0
#define LOCAL_SIZE_0 8
#endif

#ifndef LOCAL_SIZE_1
#define LOCAL_SIZE_1 8
#endif

#define CONVO_ROW_KERNEL_NAME "Convo_Row"

#define CONVO_ROW_KERNEL \
    " \n\n __kernel void Convo_Row(const int SIZE_0, const int SIZE_1, __global float *imgDst, __global float *imgDst2, \
            __global const float *imgSrc, __global const float *rKernel, __global const float *rKernel2) {  \
        int ix = get_global_id(1);  \
        int iy = get_global_id(0);  \
 \
	float sum = 0;  \
        float sum2 = 0;  \
        int d = 0;  \
 \
        if ( (ix >= SIZE_0) || (iy >= SIZE_1) ) {  \
		return;  \
	}  \
 \
        for (int k = -rKernelRadius; k <= rKernelRadius; k++) {  \
            d = ix + k;  \
            if ((d >= 0) && (d < SIZE_0)) {  \
                sum += imgSrc[d * SIZE_1 + iy] * rKernel[rKernelRadius + k];  \
                sum2 += imgSrc[d * SIZE_1 + iy] * rKernel2[rKernelRadius + k];  \
            }  \
        }  \
        imgDst[ix * SIZE_1 + iy] = sum;  \
        imgDst2[ix * SIZE_1 + iy] = sum2;  \
    } \n\n "

#define CONVO_COL_KERNEL_NAME "Convo_Col"

#define CONVO_COL_KERNEL \
    " \n\n __kernel void Convo_Col(const int SIZE_0, const int SIZE_1, __global float *imgDst, \
            __global const float *imgSrc, __global const float *cKernel) {  \
        int ix = get_global_id(1);  \
        int iy = get_global_id(0);  \
 \
	float sum = 0;n\
        int d = 0;n\
 \
        if ( (ix >= SIZE_0) || (iy >= SIZE_1) ) {  \
		return;  \
	}  \
 \
        for (int k = -cKernelRadius; k <= cKernelRadius; k++) {  \
            d = iy + k;  \
            if ((d >= 0) && (d < SIZE_1)) {  \
                sum += imgSrc[ix * SIZE_1 + d] * cKernel[cKernelRadius + k];  \
            }  \
        }  \
        imgDst[ix * SIZE_1 + iy] = sum * sum;  \
    } \n\n "

#define SAXPY_SQRT_KERNEL_NAME "Saxpy_Sqrt"

#define SAXPY_SQRT_KERNEL \
    " \n\n __kernel void Saxpy_Sqrt(const int SIZE_0, const int SIZE_1, \
            __global float *matrix_x, __global float *matrix_y) {  \
        int ix = get_global_id(1);  \
        int iy = get_global_id(0);  \
 \
        if ( (ix >= SIZE_0) || (iy >= SIZE_1) ) {  \
			return;  \
		}  \
 \
        matrix_y[ix * SIZE_1 + iy] = sqrt(alpha * matrix_x[ix * SIZE_1 + iy] + matrix_y[ix * SIZE_1 + iy]);  \
    } \n\n "


double main_clock;
double exec_clock;

int main (int argc, char *argv[]) {
	main_clock = omp_get_wtime();

	if (argc < 5) {
		fprintf(stderr, "Usage: %s <size_0> <size_1> <numIter> <platform> <device>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	int i, j, k;

	cl_int SIZE_0 = atoi(argv[1]);
    cl_int SIZE_1 = atoi(argv[2]);
	cl_int N_ITER = atoi(argv[3]);
	cl_int PLATFORM = atoi(argv[4]);
	cl_int DEVICE = atoi(argv[5]);

	cl_int err;

	size_t MATRIX_SIZE = sizeof(cl_float) * SIZE_0 * SIZE_1;

    char *kernel_raw_convo_row = CONVO_ROW_KERNEL;
	char *kernel_raw_convo_col = CONVO_COL_KERNEL;
    char *kernel_raw_saxpy_sqrt = SAXPY_SQRT_KERNEL;

	char *kernel_raw = (char *)malloc((strlen(kernel_raw_convo_row) + strlen(kernel_raw_convo_col) + strlen(kernel_raw_saxpy_sqrt) + 300) * sizeof(char));
	sprintf(&kernel_raw[0], " #define cKernelRadius %d \n#define rKernelRadius %d \n#define alpha 1.0\n", cKernelRadius, rKernelRadius);
	strcat(kernel_raw, kernel_raw_convo_row);
	strcat(kernel_raw, kernel_raw_convo_col);
    strcat(kernel_raw, kernel_raw_saxpy_sqrt);

	size_t kernel_size = strlen(kernel_raw);

	size_t local_size[2];
	size_t global_size[2];

	double result;
	double sum;

	double  execClock;
	double  totalClock;

	cl_platform_id platform_id;
	cl_device_id device_id;

	cl_context context;
	cl_program program;
	cl_kernel kernel_convo_row;
	cl_kernel kernel_convo_col;
	cl_kernel kernel_saxpy;
	cl_command_queue_properties properties;

	cl_command_queue main_command_queue;
	cl_command_queue write_command_queue;
	cl_command_queue read_command_queue;

	cl_mem mem_img_orig;
	cl_mem mem_pinned_img_orig;
	cl_float *p_pinned_img_orig;

	cl_mem mem_img_gx;
	cl_mem mem_pinned_img_gx;
	cl_float *p_pinned_img_gx;

	cl_mem mem_img_gy;
	cl_mem mem_pinned_img_gy;
	cl_float *p_pinned_img_gy;

	cl_mem mem_kernel_gxc;
	cl_mem mem_pinned_kernel_gxc;
	cl_float *p_pinned_kernel_gxc;

	cl_mem mem_kernel_gxr;
	cl_mem mem_pinned_kernel_gxr;
	cl_float *p_pinned_kernel_gxr;

	cl_mem mem_img_temp;
	cl_mem mem_img_temp2;

	cl_event event_write;
	cl_event event_read;

	cl_event event_convo_row;
	cl_event event_convo_col1;
    cl_event event_convo_col2;
    cl_event event_saxpy;

    cl_event event_default;
	cl_event event_wait_list[5];

    cl_platform_id *p_platforms = (cl_platform_id *)malloc((PLATFORM + 1) * sizeof(cl_platform_id));
	clGetPlatformIDs(PLATFORM + 1, p_platforms, NULL);
    platform_id = p_platforms[PLATFORM];
	free(p_platforms);

    cl_device_id *p_devices = (cl_device_id *)malloc((DEVICE + 1) * sizeof(cl_device_id));
    clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, DEVICE + 1, p_devices, NULL);
    device_id = p_devices[DEVICE];
    free(p_devices);

	cl_context_properties context_properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_id, 0 };
	context = clCreateContext(context_properties, 1, &device_id, NULL, NULL, &err);

	main_command_queue = clCreateCommandQueue(context, device_id, 0, &err);
	read_command_queue = clCreateCommandQueue(context, device_id, 0, &err);
	write_command_queue = clCreateCommandQueue(context, device_id, 0, &err);

	mem_pinned_kernel_gxc = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, sizeof(cl_float) * 3, NULL, &err);
	p_pinned_kernel_gxc = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_kernel_gxc, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, sizeof(cl_float) * 3, 0, NULL, NULL, &err);
	mem_kernel_gxc = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_float) * 3, NULL, &err);
	
	mem_pinned_kernel_gxr = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, sizeof(cl_float) * 3, NULL, &err);
	p_pinned_kernel_gxr = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_kernel_gxr, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, sizeof(cl_float) * 3, 0, NULL, NULL, &err);
	mem_kernel_gxr = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_float) * 3, NULL, &err);

	mem_pinned_img_orig = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	p_pinned_img_orig = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_img_orig, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	mem_img_orig = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);
	
	mem_pinned_img_gx = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	p_pinned_img_gx = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_img_gx, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	mem_img_gx = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);

	mem_pinned_img_gy = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	p_pinned_img_gy = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_img_gy, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	mem_img_gy = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);

	mem_img_temp = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_HOST_NO_ACCESS, MATRIX_SIZE, NULL, &err);
	mem_img_temp2 = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_HOST_NO_ACCESS, MATRIX_SIZE, NULL, &err);
	
	program = clCreateProgramWithSource(context, 1, (const char **)(&kernel_raw), (const size_t *)(&kernel_size), &err);

	clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
	
	kernel_convo_row = clCreateKernel(program, CONVO_ROW_KERNEL_NAME, &err);
	kernel_convo_col = clCreateKernel(program, CONVO_COL_KERNEL_NAME, &err);
	kernel_saxpy = clCreateKernel(program, SAXPY_SQRT_KERNEL_NAME, &err);

	event_default = clCreateUserEvent(context, &err);

	clSetUserEventStatus(event_default, CL_COMPLETE);

	event_convo_row = event_default;
    event_convo_col1 = event_default;
    event_convo_col2 = event_default;
    event_saxpy = event_default;
    event_read = event_default;

    clRetainEvent(event_convo_row);
    clRetainEvent(event_convo_col1);
    clRetainEvent(event_convo_col2);
    clRetainEvent(event_saxpy);
    clRetainEvent(event_read);

    clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);

	local_size[0] = LOCAL_SIZE_0;
	local_size[1] = LOCAL_SIZE_1;

	global_size[0] = SIZE_1;
    	global_size[1] = SIZE_0;

	if ((global_size[0] % local_size[0]) != 0) {
		global_size[0] += (local_size[0] - (global_size[0] % local_size[0]));
	}

	if ((global_size[1] % local_size[1]) != 0) {
		global_size[1] += (local_size[1] - (global_size[1] % local_size[1]));
	}

    for (i = 0; i < SIZE_0; i++) {
        for (j = 0; j < SIZE_1; j++) {
            p_pinned_img_orig[i * SIZE_1 + j] = 123.123;
        }
    }

    p_pinned_kernel_gxc[0] = 1;
    p_pinned_kernel_gxc[1] = 2;
    p_pinned_kernel_gxc[2] = 1;

    p_pinned_kernel_gxr[0] = -1;
    p_pinned_kernel_gxr[1] = 0;
    p_pinned_kernel_gxr[2] = 1;

	clSetKernelArg(kernel_convo_row, 0, sizeof(cl_int), &SIZE_0);
	clSetKernelArg(kernel_convo_row, 1, sizeof(cl_int), &SIZE_1);
	clSetKernelArg(kernel_convo_row, 2, sizeof(cl_mem), &mem_img_temp);
	clSetKernelArg(kernel_convo_row, 3, sizeof(cl_mem), &mem_img_temp2);
	clSetKernelArg(kernel_convo_row, 4, sizeof(cl_mem), &mem_img_orig);
	clSetKernelArg(kernel_convo_row, 5, sizeof(cl_mem), &mem_kernel_gxr);
	clSetKernelArg(kernel_convo_row, 6, sizeof(cl_mem), &mem_kernel_gxc);

    clSetKernelArg(kernel_saxpy, 0, sizeof(cl_int), &SIZE_0);
	clSetKernelArg(kernel_saxpy, 1, sizeof(cl_int), &SIZE_1);
	clSetKernelArg(kernel_saxpy, 2, sizeof(cl_mem), &mem_img_gx);
	clSetKernelArg(kernel_saxpy, 3, sizeof(cl_mem), &mem_img_gy);

	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);
	
	clEnqueueWriteBuffer(write_command_queue, mem_img_orig, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_orig, 0, NULL, NULL);
	clFlush(write_command_queue);

	clEnqueueWriteBuffer(write_command_queue, mem_kernel_gxc, CL_FALSE, 0, sizeof(cl_float) * 3, (void *)p_pinned_kernel_gxc, 0, NULL, NULL);
	clFlush(write_command_queue);

	clEnqueueWriteBuffer(write_command_queue, mem_kernel_gxr, CL_FALSE, 0, sizeof(cl_float) * 3, (void *)p_pinned_kernel_gxr, 0, NULL, &event_write);
	clFlush(write_command_queue);

    clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);
	
    exec_clock = omp_get_wtime();
    
    for (i = 0; i < N_ITER; i++) {
        clReleaseEvent(event_convo_row);
        clEnqueueNDRangeKernel(main_command_queue, kernel_convo_row, 2, NULL, global_size, local_size, 1, &event_write, &event_convo_row);
        clFlush(main_command_queue);

        clReleaseEvent(event_write);
		clEnqueueWriteBuffer(write_command_queue, mem_img_orig, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_orig, 1, &event_convo_row, &event_write);
		clFlush(write_command_queue);
        
		clSetKernelArg(kernel_convo_col, 0, sizeof(cl_int), &SIZE_0);
		clSetKernelArg(kernel_convo_col, 1, sizeof(cl_int), &SIZE_1);
		clSetKernelArg(kernel_convo_col, 2, sizeof(cl_mem), &mem_img_gx);
		clSetKernelArg(kernel_convo_col, 3, sizeof(cl_mem), &mem_img_temp);
		clSetKernelArg(kernel_convo_col, 4, sizeof(cl_mem), &mem_kernel_gxc);

        clReleaseEvent(event_convo_col1);
        clEnqueueNDRangeKernel(main_command_queue, kernel_convo_col, 2, NULL, global_size, local_size, 0, NULL, &event_convo_col1);
        clFlush(main_command_queue);

    	clSetKernelArg(kernel_convo_col, 0, sizeof(cl_int), &SIZE_0);
    	clSetKernelArg(kernel_convo_col, 1, sizeof(cl_int), &SIZE_1);
    	clSetKernelArg(kernel_convo_col, 2, sizeof(cl_mem), &mem_img_gy);
    	clSetKernelArg(kernel_convo_col, 3, sizeof(cl_mem), &mem_img_temp2);
    	clSetKernelArg(kernel_convo_col, 4, sizeof(cl_mem), &mem_kernel_gxr);

        clReleaseEvent(event_convo_col2);
        clEnqueueNDRangeKernel(main_command_queue, kernel_convo_col, 2, NULL, global_size, local_size, 1, &event_read, &event_convo_col2);
        clFlush(main_command_queue);

        clReleaseEvent(event_saxpy);
        clEnqueueNDRangeKernel(main_command_queue, kernel_saxpy, 2, NULL, global_size, local_size, 0, NULL, &event_saxpy);
        clFlush(main_command_queue);

		clReleaseEvent(event_read);
		clEnqueueReadBuffer(read_command_queue, mem_img_gy, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_gy, 1, &event_saxpy, &event_read);
		clFlush(read_command_queue);
    }

	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);

	exec_clock = omp_get_wtime() - exec_clock ;

	sum = result = 0;
	for (i = 0; i < SIZE_0; i++) {
		for (j = 0; j < SIZE_1; j++) {
			sum += pow(p_pinned_img_gy[i * SIZE_1 + j], 2);
		}
	}
	result = sqrt(sum);
 
	printf("\n ---------------------- RESULT ---------------------- \n");
	printf("\n Sum: %lf", sum);
	printf("\n Result: %lf", result);
	printf("\n\n ---------------------------------------------------- \n");

	clReleaseEvent(event_default);
	clReleaseEvent(event_write);
	clReleaseEvent(event_read);
	clReleaseEvent(event_convo_row);
	clReleaseEvent(event_convo_col1);
    clReleaseEvent(event_convo_col2);
    clReleaseEvent(event_saxpy);

	clEnqueueUnmapMemObject(main_command_queue, mem_pinned_img_orig, p_pinned_img_orig, 0, NULL, NULL);
    clEnqueueUnmapMemObject(main_command_queue, mem_pinned_img_gx, p_pinned_img_gx, 0, NULL, NULL);
    clEnqueueUnmapMemObject(main_command_queue, mem_pinned_img_gy, p_pinned_img_gy, 0, NULL, NULL);
	clEnqueueUnmapMemObject(main_command_queue, mem_pinned_kernel_gxc, p_pinned_kernel_gxc, 0, NULL, NULL);
	clEnqueueUnmapMemObject(main_command_queue, mem_pinned_kernel_gxr, p_pinned_kernel_gxr, 0, NULL, NULL);
	
	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clReleaseMemObject(mem_pinned_img_orig);
    clReleaseMemObject(mem_img_orig);

    clReleaseMemObject(mem_pinned_img_gx);
    clReleaseMemObject(mem_img_gx);

    clReleaseMemObject(mem_pinned_img_gy);
    clReleaseMemObject(mem_img_gy);

    clReleaseMemObject(mem_pinned_kernel_gxc);
    clReleaseMemObject(mem_kernel_gxc);

    clReleaseMemObject(mem_pinned_kernel_gxr);
    clReleaseMemObject(mem_kernel_gxr);

    clReleaseMemObject(mem_img_temp);
    clReleaseMemObject(mem_img_temp2) 	

	clReleaseKernel(kernel_convo_row);
    clReleaseKernel(kernel_convo_col);
    clReleaseKernel(kernel_saxpy);

	clReleaseProgram(program);
	
	clReleaseCommandQueue(main_command_queue);
	clReleaseCommandQueue(read_command_queue);
	clReleaseCommandQueue(write_command_queue);
	
	clReleaseContext(context);

	main_clock = omp_get_wtime() - main_clock ;

	printf("\n ---------------------- TIMERS ---------------------- \n");
	printf("Clock main: %lf\n", main_clock);
	printf("Clock exec: %lf\n", exec_clock);
	printf("\n\n ---------------------------------------------------- \n");
	
	return 0;
}

