#include "Ctrl.h"
#include <math.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>

#define IMAD(a, b, c) (mul24((a), (b)) + (c))
#define SEED 6834723
#define cKernelRadius 1
#define rKernelRadius 1
#define alpha 1.0

/*Tipo de datos*/
hit_tileNewType(float);
hit_ktileNewType(float);

/*Tiempos*/
double main_clock;
double exec_clock;

CTRL_KERNEL_CHAR(convolutionRowsKernel, MANUAL, 2, 128, 2);
CTRL_KERNEL_CHAR(convolutionColumnsKernel, MANUAL, 2, 128, 2);
CTRL_KERNEL_CHAR(kernel_saxpy_sqrt, MANUAL, 2, 128, 2);


CTRL_KERNEL(convolutionRowsKernel, OPENCL_GPU, KHitTile_float imgDst, KHitTile_float imgDst2,
		KHitTile_float imgSrc, KHitTile_float rKernel, KHitTile_float rKernel2,
{
	float sum = 0;
	float sum2 = 0;
	int d = 0;
	for (int k = -rKernelRadius; k <= rKernelRadius; k++) {
		d = thread_id.x + k;
		if ((d >= 0) && (d < imgSrc.card[0])) {
			sum += (hit_tileElemAtNoStride(imgSrc, 2, d, thread_id.y) * hit_tileElemAtNoStride(rKernel, 1, rKernelRadius + k));
			sum2 += (hit_tileElemAtNoStride(imgSrc, 2, d, thread_id.y) * hit_tileElemAtNoStride(rKernel2, 1, rKernelRadius + k));
		}
	}
	hit_tileElemAtNoStride(imgDst, 2, thread_id.x, thread_id.y) = sum; \n
	hit_tileElemAtNoStride(imgDst2, 2, thread_id.x, thread_id.y) = sum2; \n
});


CTRL_KERNEL(convolutionColumnsKernel, OPENCL_GPU, KHitTile_float imgDst,
		KHitTile_float imgSrc, KHitTile_float cKernel,
{
	float sum = 0;
	int d = 0;
	for (int k = -cKernelRadius; k <= cKernelRadius; k++) {
		d = thread_id.y + k;
		if ((d >= 0) && (d < imgSrc.card[1])) {
			sum += (hit_tileElemAtNoStride(imgSrc, 2, thread_id.x, d) * hit_tileElemAtNoStride(cKernel, 1, cKernelRadius + k));
		}
	}
	hit_tileElemAtNoStride(imgDst, 2, thread_id.x, thread_id.y) = sum * sum;
});


CTRL_KERNEL(kernel_saxpy_sqrt, OPENCL_GPU, KHitTile_float matrix_x,
		KHitTile_float matrix_y,
{
	hit_tileElemAtNoStride(matrix_y, 2, thread_id.x, thread_id.y) =
	sqrt(alpha *
	hit_tileElemAtNoStride(matrix_x, 2, thread_id.x, thread_id.y)
	+ hit_tileElemAtNoStride(matrix_y, 2, thread_id.x, thread_id.y));
});


CTRL_KERNEL_PROTO(convolutionRowsKernel, 1, OPENCL_GPU,
		5, 
		OUT, HitTile_float, imgDst, 
		OUT, HitTile_float, imgDst2,
		IN, HitTile_float, imgSrc, 
		IN, HitTile_float, rKernel,
		IN, HitTile_float, rKernel2);

CTRL_KERNEL_PROTO(convolutionColumnsKernel, 1, OPENCL_GPU, 
		3, 
		OUT, HitTile_float, imgDst, 
		IN, HitTile_float, imgSrc, 
		IN, HitTile_float, cKernel);

CTRL_KERNEL_PROTO(kernel_saxpy_sqrt, 1, OPENCL_GPU, 
		2, 
		IN, HitTile_float, matrix_x, 
		IO, HitTile_float, matrix_y);


int main(int argc, char *argv[]) {
	main_clock = omp_get_wtime();

	if (argc != 6) {
		fprintf(stderr, "Usage: %s <matrixSize> <matrixSize> <iter> <platform ><device>\n\n",
				argv[0]);
		exit(EXIT_FAILURE);
	}


	int SIZE1 = atoi(argv[1]);
	int SIZE2 = atoi(argv[2]);
	int N_ITER = atoi(argv[3]);
	int PLATFORM = atoi(argv[4]);
	int DEVICE = atoi(argv[5]);

	HitTile_float h_KernelGxC, h_KernelGxR, imgOriginal, imgTemp, imgTemp2, imgGx, imgGy;

	hit_tileDomain(&h_KernelGxC, float, 1, 3);
	hit_tileDomain(&h_KernelGxR, float, 1, 3);
	hit_tileDomain(&imgOriginal, float, 2, SIZE1, SIZE2);
	hit_tileDomain(&imgTemp, float, 2, SIZE1, SIZE2);
	hit_tileDomain(&imgTemp2, float, 2, SIZE1, SIZE2);
	hit_tileDomain(&imgGx, float, 2, SIZE1, SIZE2);
	hit_tileDomain(&imgGy, float, 2, SIZE1, SIZE2);

	Ctrl_Thread threads;
	Ctrl_ThreadInit(threads, 2, SIZE1, SIZE2);

	__ctrl_block__(1)
	{
		Ctrl ctrl;
		Ctrl_Create(&ctrl, CTRL_TYPE_OPENCL_GPU, CTRL_POLICY_ASYNC, DEVICE, &PLATFORM);

		Ctrl_AllocPinned(&ctrl, (HitTile *)&h_KernelGxC);
		Ctrl_AllocPinned(&ctrl, (HitTile *)&h_KernelGxR);
		Ctrl_AllocPinned(&ctrl, (HitTile *)&imgOriginal);
		Ctrl_AllocPinned(&ctrl, (HitTile *)&imgGx);
		Ctrl_AllocPinned(&ctrl, (HitTile *)&imgGy);

		// srand48(SEED);

		for (unsigned int i = 0; i < SIZE1; i++) {
			for (unsigned int j = 0; j < SIZE2; j++) {
				hit_tileElemAt(imgOriginal, 2, i, j) = 123.123; //(float)(drand48());
			}
		}

		hit_tileElemAt(h_KernelGxC, 1, 0) = 1;
		hit_tileElemAt(h_KernelGxC, 1, 1) = 2;
		hit_tileElemAt(h_KernelGxC, 1, 2) = 1;
		hit_tileElemAt(h_KernelGxR, 1, 0) = -1;
		hit_tileElemAt(h_KernelGxR, 1, 1) = 0;
		hit_tileElemAt(h_KernelGxR, 1, 2) = 1;

		Ctrl_Attach(&ctrl, (HitTile *)&h_KernelGxC);
		Ctrl_Attach(&ctrl, (HitTile *)&h_KernelGxR);
		Ctrl_Attach(&ctrl, (HitTile *)&imgOriginal);
		Ctrl_Attach(&ctrl, (HitTile *)&imgGx);
		Ctrl_Attach(&ctrl, (HitTile *)&imgGy);
		Ctrl_CreateInternal(&ctrl, (HitTile *)&imgTemp);
		Ctrl_CreateInternal(&ctrl, (HitTile *)&imgTemp2);

		Ctrl_GlobalSync(&ctrl);
		exec_clock = omp_get_wtime();

		for (int iter = 0; iter < N_ITER; iter++) {
			Ctrl_Launch(&ctrl, convolutionRowsKernel, threads, imgTemp, imgTemp2, imgOriginal, h_KernelGxR, h_KernelGxC);
			Ctrl_MoveTo(&ctrl, (HitTile *)&imgOriginal);
			Ctrl_Launch(&ctrl, convolutionColumnsKernel, threads, imgGx, imgTemp, h_KernelGxC);
			Ctrl_Launch(&ctrl, convolutionColumnsKernel, threads, imgGy, imgTemp2, h_KernelGxR);
			Ctrl_Launch(&ctrl, kernel_saxpy_sqrt, threads, imgGx, imgGy); 
			Ctrl_MoveFrom(&ctrl, (HitTile *)&imgGy);
		}

		Ctrl_GlobalSync(&ctrl);
		exec_clock = omp_get_wtime() - exec_clock;

		Ctrl_Detach(&ctrl, (HitTile *)&h_KernelGxC);
		Ctrl_Detach(&ctrl, (HitTile *)&h_KernelGxR);
		Ctrl_Detach(&ctrl, (HitTile *)&imgOriginal);
		Ctrl_Detach(&ctrl, (HitTile *)&imgGy);
		Ctrl_Detach(&ctrl, (HitTile *)&imgGx);
		Ctrl_DestroyInternal(&ctrl, (HitTile *)&imgTemp);
		Ctrl_DestroyInternal(&ctrl, (HitTile *)&imgTemp2);

		double resultado = 0, suma = 0;
		for (int i = 0; i < SIZE1; i++) {
			for (int j = 0; j < SIZE2; j++) {
				suma += pow(hit_tileElemAt(imgGy, 2, i, j), 2);
			}
		}

		printf("\n ----------------------- NORM ----------------------- \n\n");
		fflush(stdout);
		printf(" Acumulated sum: %lf \n", suma);
		fflush(stdout);

		resultado = sqrt(suma);
		printf(" Result: %lf \n", resultado);
		fflush(stdout);

		printf("\n ---------------------------------------------------- \n");
		fflush(stdout);

		Ctrl_FreePinned(&ctrl, (HitTile *)&h_KernelGxC);
		Ctrl_FreePinned(&ctrl, (HitTile *)&h_KernelGxR);
		Ctrl_FreePinned(&ctrl, (HitTile *)&imgOriginal);
		Ctrl_FreePinned(&ctrl, (HitTile *)&imgGx);
		Ctrl_FreePinned(&ctrl, (HitTile *)&imgGy);

		Ctrl_Destroy(&ctrl);
	}

	main_clock = omp_get_wtime() - main_clock;

	printf("\n ----------------------- TIME ----------------------- \n\n");
	printf(" Clock main: %lf\n", main_clock);
	printf(" Clock exec: %lf\n", exec_clock);
	printf("\n ---------------------------------------------------- \n");

	return 0;
}
