#define CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <CL/cl.h>
#include <math.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define IMAD(a, b, c) (mul24((a), (b)) + (c))
#define SEED 6834723
#define cKernelRadius 1
#define rKernelRadius 1
#define alpha 1.0

#ifndef LOCAL_SIZE_0
#define LOCAL_SIZE_0 8
#endif

#ifndef LOCAL_SIZE_1
#define LOCAL_SIZE_1 8
#endif

#ifdef _CTRL_EXAMPLES_OPENCL_GPU_ERROR_CHECK_
	#include <assert.h>
	
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_DEBUG_
		#define OPENCL_ASSERT_OP( operation ) \
			err = operation; \
			printf("error: %d\n", err); \
			assert(err == CL_SUCCESS);

		#define OPENCL_ASSERT_ERROR( err ) \
			printf("error: %d\n", err); \
			assert(err == CL_SUCCESS)

	#else 
		#define OPENCL_ASSERT_OP( operation ) \
			assert(operation == CL_SUCCESS);

		#define OPENCL_ASSERT_ERROR( err ) \
			assert(err == CL_SUCCESS)
	#endif
#else
	#define OPENCL_ASSERT_OP( operation ) operation
	#define OPENCL_ASSERT_ERROR(err)
#endif



#define CONVO_ROW_KERNEL_NAME "Convo_Row"

#define CONVO_ROW_KERNEL \
    " \n\n __kernel void Convo_Row(const int SIZE_0, const int SIZE_1, __global float *imgDst, __global float *imgDst2, \
            __global const float *imgSrc, __global const float *rKernel, __global const float *rKernel2) {  \
        int ix = get_global_id(1);  \
        int iy = get_global_id(0);  \
 \
	float sum = 0;  \
        float sum2 = 0;  \
        int d = 0;  \
 \
        if ( (ix >= SIZE_0) || (iy >= SIZE_1) ) {  \
		return;  \
	}  \
 \
        for (int k = -rKernelRadius; k <= rKernelRadius; k++) {  \
            d = ix + k;  \
            if ((d >= 0) && (d < SIZE_0)) {  \
                sum += imgSrc[d * SIZE_1 + iy] * rKernel[rKernelRadius + k];  \
                sum2 += imgSrc[d * SIZE_1 + iy] * rKernel2[rKernelRadius + k];  \
            }  \
        }  \
        imgDst[ix * SIZE_1 + iy] = sum;  \
        imgDst2[ix * SIZE_1 + iy] = sum2;  \
    } \n\n "

#define CONVO_COL_KERNEL_NAME "Convo_Col"

#define CONVO_COL_KERNEL \
    " \n\n __kernel void Convo_Col(const int SIZE_0, const int SIZE_1, __global float *imgDst, \
            __global const float *imgSrc, __global const float *cKernel) {  \
        int ix = get_global_id(1);  \
        int iy = get_global_id(0);  \
 \
	float sum = 0;n\
        int d = 0;n\
 \
        if ( (ix >= SIZE_0) || (iy >= SIZE_1) ) {  \
		return;  \
	}  \
 \
        for (int k = -cKernelRadius; k <= cKernelRadius; k++) {  \
            d = iy + k;  \
            if ((d >= 0) && (d < SIZE_1)) {  \
                sum += imgSrc[ix * SIZE_1 + d] * cKernel[cKernelRadius + k];  \
            }  \
        }  \
        imgDst[ix * SIZE_1 + iy] = sum * sum;  \
    } \n\n "

#define SAXPY_SQRT_KERNEL_NAME "Saxpy_Sqrt"

#define SAXPY_SQRT_KERNEL \
    " \n\n __kernel void Saxpy_Sqrt(const int SIZE_0, const int SIZE_1, \
            __global float *matrix_x, __global float *matrix_y) {  \
        int ix = get_global_id(1);  \
        int iy = get_global_id(0);  \
 \
        if ( (ix >= SIZE_0) || (iy >= SIZE_1) ) {  \
			return;  \
		}  \
 \
        matrix_y[ix * SIZE_1 + iy] = sqrt(alpha * matrix_x[ix * SIZE_1 + iy] + matrix_y[ix * SIZE_1 + iy]);  \
    } \n\n "


double main_clock;
double exec_clock;

int main (int argc, char *argv[]) {
	main_clock = omp_get_wtime();
	/* ARGUMENTS */

	if (argc < 5) {
		fprintf(stderr, "Usage: %s <size_0> <size_1> <numIter> <platform> <device>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	printf("init\n");

	int i, j, k;

	cl_int SIZE_0 = atoi(argv[1]);
    	cl_int SIZE_1 = atoi(argv[2]);
	cl_int N_ITER = atoi(argv[3]);
	cl_int PLATFORM = atoi(argv[4]);
	cl_int DEVICE = atoi(argv[5]);

	size_t MATRIX_SIZE = sizeof(cl_float) * SIZE_0 * SIZE_1;

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
			printf("\n ----------------------- ARGS ----------------------- \n");
			printf("\n SIZE_0 x SIZE_1: %d x %d", SIZE_0, SIZE_1);
			printf("\n N_ITER: %d", N_ITER);
			printf("\n PLATFORM: %d", PLATFORM);
			printf("\n DEVICE: %d", DEVICE);
			#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
				printf("\n POLICY ASYNC");
			#else // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
				printf("\n POLICY SYNC");
			#endif // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

	/* VARIABLES */

	cl_int err;

    char *kernel_raw_convo_row = CONVO_ROW_KERNEL;
	char *kernel_raw_convo_col = CONVO_COL_KERNEL;
    char *kernel_raw_saxpy_sqrt = SAXPY_SQRT_KERNEL;

	char *kernel_raw = (char *)malloc((strlen(kernel_raw_convo_row) + strlen(kernel_raw_convo_col) + strlen(kernel_raw_saxpy_sqrt) + 300) * sizeof(char));
	sprintf(&kernel_raw[0], " #define cKernelRadius %d \n#define rKernelRadius %d \n#define alpha 1.0\n", cKernelRadius, rKernelRadius);
	strcat(kernel_raw, kernel_raw_convo_row);
	strcat(kernel_raw, kernel_raw_convo_col);
    strcat(kernel_raw, kernel_raw_saxpy_sqrt);


	size_t kernel_size = strlen(kernel_raw);

	size_t local_size[2];
	size_t global_size[2];

	double result;
	double sum;

	double  execClock;
	double  totalClock;

	cl_platform_id platform_id;
	cl_device_id device_id;

	cl_context context;
	cl_program program;
	cl_kernel kernel_convo_row;
    	cl_kernel kernel_convo_col;
    	// cl_kernel kernel_convo_col1;
    	// cl_kernel kernel_convo_col2;
    	cl_kernel kernel_saxpy;
	cl_command_queue_properties properties;

	cl_command_queue main_command_queue;
	cl_command_queue write_command_queue;
	cl_command_queue read_command_queue;

	cl_mem mem_img_orig;
	cl_mem mem_pinned_img_orig;
    	cl_float *p_pinned_img_orig;

    	cl_mem mem_img_gx;
	cl_mem mem_pinned_img_gx;
    	cl_float *p_pinned_img_gx;

	cl_mem mem_img_gy;
	cl_mem mem_pinned_img_gy;
    	cl_float *p_pinned_img_gy;

	cl_mem mem_kernel_gxc;
	cl_mem mem_pinned_kernel_gxc;
    	cl_float *p_pinned_kernel_gxc;

    	cl_mem mem_kernel_gxr;
	cl_mem mem_pinned_kernel_gxr;
    	cl_float *p_pinned_kernel_gxr;

    	cl_mem mem_img_temp;
    	cl_mem mem_img_temp2;


	cl_event event_write;
	cl_event event_read;

	cl_event event_convo_row;
	cl_event event_convo_col1;
    cl_event event_convo_col2;
    cl_event event_saxpy;

    cl_event event_default;

	cl_event event_wait_list[5];

    cl_platform_id *p_platforms = (cl_platform_id *)malloc((PLATFORM + 1) * sizeof(cl_platform_id));
	OPENCL_ASSERT_OP( clGetPlatformIDs(PLATFORM + 1, p_platforms, NULL) );
    platform_id = p_platforms[PLATFORM];
	free(p_platforms);

    cl_device_id *p_devices = (cl_device_id *)malloc((DEVICE + 1) * sizeof(cl_device_id));
    OPENCL_ASSERT_OP( clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, DEVICE + 1, p_devices, NULL) );
    device_id = p_devices[DEVICE];
    free(p_devices);


	cl_context_properties context_properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_id, 0 };
	context = clCreateContext(context_properties, 1, &device_id, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	properties = 0;

	main_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
	OPENCL_ASSERT_ERROR( err );
	
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		read_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
		OPENCL_ASSERT_ERROR( err );
		write_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
		OPENCL_ASSERT_ERROR( err );
	#endif


	mem_pinned_kernel_gxc = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, sizeof(cl_float) * 3, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	p_pinned_kernel_gxc = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_kernel_gxc, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, sizeof(cl_float) * 3, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	mem_kernel_gxc = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_float) * 3, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	
	mem_pinned_kernel_gxr = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, sizeof(cl_float) * 3, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	p_pinned_kernel_gxr = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_kernel_gxr, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, sizeof(cl_float) * 3, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	mem_kernel_gxr = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_float) * 3, NULL, &err);
	OPENCL_ASSERT_ERROR( err );


	mem_pinned_img_orig = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	p_pinned_img_orig = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_img_orig, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	mem_img_orig = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	
	mem_pinned_img_gx = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	p_pinned_img_gx = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_img_gx, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	mem_img_gx = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );


	mem_pinned_img_gy = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	p_pinned_img_gy = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_img_gy, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	mem_img_gy = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );


	mem_img_temp = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_HOST_NO_ACCESS, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );


	mem_img_temp2 = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_HOST_NO_ACCESS, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	
	
	program = clCreateProgramWithSource(context, 1, (const char **)(&kernel_raw), (const size_t *)(&kernel_size), &err);
	OPENCL_ASSERT_ERROR( err );

	err = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
	if (err == CL_BUILD_PROGRAM_FAILURE) {
		printf("\n\n%s\n\n", kernel_raw);
		size_t log_size;
		clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
		char *log = (char *) malloc(log_size);
		clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, log_size, log, NULL);
		printf("%s\n", log);
		free(log);
	}
	OPENCL_ASSERT_ERROR( err );
	
	kernel_convo_row = clCreateKernel(program, CONVO_ROW_KERNEL_NAME, &err);
	OPENCL_ASSERT_ERROR( err );

	kernel_convo_col = clCreateKernel(program, CONVO_COL_KERNEL_NAME, &err);
	OPENCL_ASSERT_ERROR( err );

    // kernel_convo_col1 = clCreateKernel(program, CONVO_COL_KERNEL_NAME, &err);
	// OPENCL_ASSERT_ERROR( err );

    // kernel_convo_col2 = clCreateKernel(program, CONVO_COL_KERNEL_NAME, &err);
	// OPENCL_ASSERT_ERROR( err );

	kernel_saxpy = clCreateKernel(program, SAXPY_SQRT_KERNEL_NAME, &err);
	OPENCL_ASSERT_ERROR( err );



	event_default = clCreateUserEvent(context, &err);
	OPENCL_ASSERT_ERROR( err );

	OPENCL_ASSERT_OP( clSetUserEventStatus(event_default, CL_COMPLETE) );

	event_convo_row = event_default;
    event_convo_col1 = event_default;
    event_convo_col2 = event_default;
    event_saxpy = event_default;
    event_read = event_default;

    //OPENCL_ASSERT_OP( clRetainEvent(event_default) );
    OPENCL_ASSERT_OP( clRetainEvent(event_convo_row) );
    OPENCL_ASSERT_OP( clRetainEvent(event_convo_col1) );
    OPENCL_ASSERT_OP( clRetainEvent(event_convo_col2) );
    OPENCL_ASSERT_OP( clRetainEvent(event_saxpy) );
    OPENCL_ASSERT_OP( clRetainEvent(event_read) );


    OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif












	local_size[0] = LOCAL_SIZE_0;
	local_size[1] = LOCAL_SIZE_1;

	global_size[0] = SIZE_1;
    	global_size[1] = SIZE_0;

	if ((global_size[0] % local_size[0]) != 0) {
		global_size[0] += (local_size[0] - (global_size[0] % local_size[0]));
	}

	if ((global_size[1] % local_size[1]) != 0) {
		global_size[1] += (local_size[1] - (global_size[1] % local_size[1]));
	}


	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
			printf("\n ------------------ WORK-GROUP-INFO ----------------- \n");
			printf("\n Local Size: %lu x %lu", local_size[0], local_size[1]);
			printf("\n Global Size: %lu x %lu", global_size[0], global_size[1]);
			printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

	//srand48(SEED);

    for (i = 0; i < SIZE_0; i++) {
        for (j = 0; j < SIZE_1; j++) {
            p_pinned_img_orig[i * SIZE_1 + j] = 123.123; //(float)(drand48());
        }
    }

    p_pinned_kernel_gxc[0] = 1;
    p_pinned_kernel_gxc[1] = 2;
    p_pinned_kernel_gxc[2] = 1;

    p_pinned_kernel_gxr[0] = -1;
    p_pinned_kernel_gxr[1] = 0;
    p_pinned_kernel_gxr[2] = 1;




	err  = clSetKernelArg(kernel_convo_row, 0, sizeof(cl_int), &SIZE_0);
	err |= clSetKernelArg(kernel_convo_row, 1, sizeof(cl_int), &SIZE_1);
	err |= clSetKernelArg(kernel_convo_row, 2, sizeof(cl_mem), &mem_img_temp);
	err |= clSetKernelArg(kernel_convo_row, 3, sizeof(cl_mem), &mem_img_temp2);
	err |= clSetKernelArg(kernel_convo_row, 4, sizeof(cl_mem), &mem_img_orig);
	err |= clSetKernelArg(kernel_convo_row, 5, sizeof(cl_mem), &mem_kernel_gxr);
	err |= clSetKernelArg(kernel_convo_row, 6, sizeof(cl_mem), &mem_kernel_gxc);
	OPENCL_ASSERT_ERROR( err );


    err  = clSetKernelArg(kernel_saxpy, 0, sizeof(cl_int), &SIZE_0);
	err |= clSetKernelArg(kernel_saxpy, 1, sizeof(cl_int), &SIZE_1);
	err |= clSetKernelArg(kernel_saxpy, 2, sizeof(cl_mem), &mem_img_gx);
	err |= clSetKernelArg(kernel_saxpy, 3, sizeof(cl_mem), &mem_img_gy);
	OPENCL_ASSERT_ERROR( err );


	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif


	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(write_command_queue, mem_img_orig, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_orig, 0, NULL, NULL) );
        OPENCL_ASSERT_OP( clFlush(write_command_queue) );
	#else
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(main_command_queue, mem_img_orig, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_orig, 0, NULL, NULL) );
        OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	#endif

    

    #ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(write_command_queue, mem_kernel_gxc, CL_FALSE, 0, sizeof(cl_float) * 3, (void *)p_pinned_kernel_gxc, 0, NULL, NULL) );
        OPENCL_ASSERT_OP( clFlush(write_command_queue) );
	#else
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(main_command_queue, mem_kernel_gxc, CL_FALSE, 0, sizeof(cl_float) * 3, (void *)p_pinned_kernel_gxc, 0, NULL, NULL) );
        OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	#endif


    #ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(write_command_queue, mem_kernel_gxr, CL_FALSE, 0, sizeof(cl_float) * 3, (void *)p_pinned_kernel_gxr, 0, NULL, &event_write) );
        OPENCL_ASSERT_OP( clFlush(write_command_queue) );
	#else
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(main_command_queue, mem_kernel_gxr, CL_FALSE, 0, sizeof(cl_float) * 3, (void *)p_pinned_kernel_gxr, 0, NULL, &event_write) );
        OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	#endif


    OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif
    exec_clock = omp_get_wtime();
    
    for (i = 0; i < N_ITER; i++) {
        OPENCL_ASSERT_OP( clReleaseEvent(event_convo_row) );
        OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_convo_row, 2, NULL, global_size, local_size, 1, &event_write, &event_convo_row) );
        OPENCL_ASSERT_OP( clFlush(main_command_queue) );


        OPENCL_ASSERT_OP( clReleaseEvent(event_write) );
        #ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
            OPENCL_ASSERT_OP( clEnqueueWriteBuffer(write_command_queue, mem_img_orig, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_orig, 1, &event_convo_row, &event_write) );
            OPENCL_ASSERT_OP( clFlush(write_command_queue) );
        #else
            OPENCL_ASSERT_OP( clEnqueueWriteBuffer(main_command_queue, mem_img_orig, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_orig, 1, &event_convo_row, &event_write) );
            OPENCL_ASSERT_OP( clFlush(main_command_queue) );
        #endif

    err  = clSetKernelArg(kernel_convo_col, 0, sizeof(cl_int), &SIZE_0);
    err |= clSetKernelArg(kernel_convo_col, 1, sizeof(cl_int), &SIZE_1);
    err |= clSetKernelArg(kernel_convo_col, 2, sizeof(cl_mem), &mem_img_gx);
    err |= clSetKernelArg(kernel_convo_col, 3, sizeof(cl_mem), &mem_img_temp);
    err |= clSetKernelArg(kernel_convo_col, 4, sizeof(cl_mem), &mem_kernel_gxc);
    OPENCL_ASSERT_ERROR( err );

        OPENCL_ASSERT_OP( clReleaseEvent(event_convo_col1) );
        OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_convo_col, 2, NULL, global_size, local_size, 0, NULL, &event_convo_col1) );
        OPENCL_ASSERT_OP( clFlush(main_command_queue) );

    err  = clSetKernelArg(kernel_convo_col, 0, sizeof(cl_int), &SIZE_0);
    err |= clSetKernelArg(kernel_convo_col, 1, sizeof(cl_int), &SIZE_1);
    err |= clSetKernelArg(kernel_convo_col, 2, sizeof(cl_mem), &mem_img_gy);
    err |= clSetKernelArg(kernel_convo_col, 3, sizeof(cl_mem), &mem_img_temp2);
    err |= clSetKernelArg(kernel_convo_col, 4, sizeof(cl_mem), &mem_kernel_gxr);
    OPENCL_ASSERT_ERROR( err );

        OPENCL_ASSERT_OP( clReleaseEvent(event_convo_col2) );
        OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_convo_col, 2, NULL, global_size, local_size, 1, &event_read, &event_convo_col2) );
        OPENCL_ASSERT_OP( clFlush(main_command_queue) );


        OPENCL_ASSERT_OP( clReleaseEvent(event_saxpy) );
        OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_saxpy, 2, NULL, global_size, local_size, 0, NULL, &event_saxpy) );
        OPENCL_ASSERT_OP( clFlush(main_command_queue) );


         OPENCL_ASSERT_OP( clReleaseEvent(event_read) );
        #ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
            OPENCL_ASSERT_OP( clEnqueueReadBuffer(read_command_queue, mem_img_gy, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_gy, 1, &event_saxpy, &event_read) );
            OPENCL_ASSERT_OP( clFlush(read_command_queue) );
        #else
            OPENCL_ASSERT_OP( clEnqueueReadBuffer(main_command_queue, mem_img_gy, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_img_gy, 1, &event_saxpy, &event_read) );
            OPENCL_ASSERT_OP( clFlush(main_command_queue) );
        #endif
    }


	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif

	exec_clock = omp_get_wtime() - exec_clock ;


    

	sum = result = 0;
	for (i = 0; i < SIZE_0; i++) {
		for (j = 0; j < SIZE_1; j++) {
			sum += pow(p_pinned_img_gy[i * SIZE_1 + j], 2);
		}
	}
	result = sqrt(sum);

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
		printf("%lf %lf\n", sum, result);
	#else 
		printf("\n ---------------------- RESULT ---------------------- \n");
		printf("\n Sum: %lf", sum);
		printf("\n Result: %lf", result);
		printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_



	OPENCL_ASSERT_OP( clReleaseEvent(event_default) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_write) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_read) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_convo_row) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_convo_col1) );
    OPENCL_ASSERT_OP( clReleaseEvent(event_convo_col2) );
    OPENCL_ASSERT_OP( clReleaseEvent(event_saxpy) );

	OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(main_command_queue, mem_pinned_img_orig, p_pinned_img_orig, 0, NULL, NULL) );
    OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(main_command_queue, mem_pinned_img_gx, p_pinned_img_gx, 0, NULL, NULL) );
    OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(main_command_queue, mem_pinned_img_gy, p_pinned_img_gy, 0, NULL, NULL) );
	OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(main_command_queue, mem_pinned_kernel_gxc, p_pinned_kernel_gxc, 0, NULL, NULL) );
	OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(main_command_queue, mem_pinned_kernel_gxr, p_pinned_kernel_gxr, 0, NULL, NULL) );
	
	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	OPENCL_ASSERT_OP( clReleaseMemObject(mem_pinned_img_orig) );
    OPENCL_ASSERT_OP( clReleaseMemObject(mem_img_orig) );

    OPENCL_ASSERT_OP( clReleaseMemObject(mem_pinned_img_gx) );
    OPENCL_ASSERT_OP( clReleaseMemObject(mem_img_gx) );

    OPENCL_ASSERT_OP( clReleaseMemObject(mem_pinned_img_gy) );
    OPENCL_ASSERT_OP( clReleaseMemObject(mem_img_gy) );

    OPENCL_ASSERT_OP( clReleaseMemObject(mem_pinned_kernel_gxc) );
    OPENCL_ASSERT_OP( clReleaseMemObject(mem_kernel_gxc) );

    OPENCL_ASSERT_OP( clReleaseMemObject(mem_pinned_kernel_gxr) );
    OPENCL_ASSERT_OP( clReleaseMemObject(mem_kernel_gxr) );

    OPENCL_ASSERT_OP( clReleaseMemObject(mem_img_temp) );
    OPENCL_ASSERT_OP( clReleaseMemObject(mem_img_temp2) );	

	OPENCL_ASSERT_OP( clReleaseKernel(kernel_convo_row) );
    OPENCL_ASSERT_OP( clReleaseKernel(kernel_convo_col) );
    // OPENCL_ASSERT_OP( clReleaseKernel(kernel_convo_col1) );
    // OPENCL_ASSERT_OP( clReleaseKernel(kernel_convo_col2) );
    OPENCL_ASSERT_OP( clReleaseKernel(kernel_saxpy) );

	OPENCL_ASSERT_OP( clReleaseProgram(program) );
	
	OPENCL_ASSERT_OP( clReleaseCommandQueue(main_command_queue) );
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clReleaseCommandQueue(read_command_queue) );
		OPENCL_ASSERT_OP( clReleaseCommandQueue(write_command_queue) );
	#endif
	
	OPENCL_ASSERT_OP( clReleaseContext(context) );


	main_clock = omp_get_wtime() - main_clock ;

		printf("\n ---------------------- TIMERS ---------------------- \n");
		printf("Clock main: %lf\n", main_clock);
		printf("Clock exec: %lf\n", exec_clock);
		printf("\n\n ---------------------------------------------------- \n");
	
	return 0;
}

