#define CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <CL/cl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

#ifndef LOCAL_SIZE_0
#define LOCAL_SIZE_0 8
#endif

#ifndef LOCAL_SIZE_1
#define LOCAL_SIZE_1 8
#endif

#define CELL_AUTOM_KERNEL_NAME "Cell_Autom"

#define CELL_AUTOM_KERNEL \
    "__kernel void Cell_Autom(const int SIZE, __global float *p_dest, __global const float *p_src) { \
        int row = get_global_id(1) + 1; \
		int col = get_global_id(0) + 1; \
        \
	if ( (row < SIZE - 1) && (col < SIZE - 1) ) { \
        /*if (!(row >= SIZE - 1 || col >= SIZE - 1)) { */ \
			p_dest[row * SIZE + col] = \
			( \
				p_src[(row - 1) * SIZE + col] + \
				p_src[(row + 1) * SIZE + col] + \
				p_src[row * SIZE + (col - 1)] + \
				p_src[row * SIZE + (col + 1)] \
			) / 4.0; \
		} \
    } \n\n"

double main_clock;
double exec_clock;

int main (int argc, char *argv[]) {
	main_clock = omp_get_wtime();
	
	if (argc < 5) {
		fprintf(stderr, "Usage: %s <size> <numIter> <platform> <device>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	cl_int SIZE = atoi(argv[1]);
	cl_int N_ITER = atoi(argv[2]);
	cl_int PLATFORM = atoi(argv[3]);
	cl_int DEVICE = atoi(argv[4]);

	size_t MATRIX_SIZE = sizeof(cl_float) * SIZE * SIZE;
	size_t MATRIX_TAM = SIZE * SIZE;

	int i, j, k;

	cl_int err;

    char *kernel_raw = CELL_AUTOM_KERNEL;

	size_t kernel_size = strlen(kernel_raw);

	size_t local_size[2];
	size_t global_size[2];

	double result;
	double sum;

	double  execClock;
	double  totalClock;

	cl_platform_id platform_id;
	cl_device_id device_id;

	cl_context context;
	cl_program program;
	cl_kernel kernel;
	cl_command_queue_properties properties;

	cl_command_queue main_command_queue;
	cl_command_queue write_command_queue;
	cl_command_queue read_command_queue;

	cl_mem mem_a;
	cl_mem mem_pinned_a;

	cl_mem mem_b;
	cl_mem mem_pinned_b;

	cl_float *p_pinned_a;
	cl_float *p_pinned_b;

	cl_event event_write;
	cl_event event_read;
	cl_event event_kernel;
	cl_event event_base;

	cl_event event_wait_list[3];

    cl_platform_id *p_platforms = (cl_platform_id *)malloc((PLATFORM + 1) * sizeof(cl_platform_id));
	clGetPlatformIDs(PLATFORM + 1, p_platforms, NULL);
    platform_id = p_platforms[PLATFORM];
	free(p_platforms);

    cl_device_id *p_devices = (cl_device_id *)malloc((DEVICE + 1) * sizeof(cl_device_id));
    clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, DEVICE + 1, p_devices, NULL);
    device_id = p_devices[DEVICE];
    free(p_devices);

	cl_context_properties context_properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_id, 0 };
	context = clCreateContext(context_properties, 1, &device_id, NULL, NULL, &err);

	main_command_queue = clCreateCommandQueue(context, device_id, 0, &err);
	read_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
	write_command_queue = clCreateCommandQueue(context, device_id, properties, &err);

	mem_pinned_a = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	p_pinned_a = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_a, CL_FALSE, CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	mem_a = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);

	mem_pinned_b = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	p_pinned_b = (float *)clEnqueueMapBuffer(main_command_queue, mem_pinned_b, CL_FALSE, CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	mem_b = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);
	
	program = clCreateProgramWithSource(context, 1, (const char **)(&kernel_raw), (const size_t *)(&kernel_size), &err);
	clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
	kernel = clCreateKernel(program, CELL_AUTOM_KERNEL_NAME, &err);

	local_size[0] = LOCAL_SIZE_0;
	local_size[1] = LOCAL_SIZE_1;

	global_size[0] = global_size[1] = SIZE;

	if ((global_size[0] % local_size[0]) != 0) {
		global_size[0] += (local_size[0] - (global_size[0] % local_size[0]));
	}

	if ((global_size[1] % local_size[1]) != 0) {
		global_size[1] += (local_size[1] - (global_size[1] % local_size[1]));
	}

	event_base = clCreateUserEvent(context, &err);

	clSetUserEventStatus(event_base, CL_COMPLETE);

	event_write = event_base;
	event_read = event_base;
	event_kernel = event_base;

	clRetainEvent(event_write);
	clRetainEvent(event_read);
	clRetainEvent(event_kernel);

    for (int i = 0; i < SIZE; i++) {
        for (int j = 0; j < SIZE; j++) {
            p_pinned_a[i * SIZE + j] = 1.1;
            p_pinned_b[i * SIZE + j] = 0.0;
        }
    }

    for (int j = 0; j < SIZE; j++){
        p_pinned_a[j] = 2.1;
        p_pinned_a[(SIZE - 1)* SIZE + j] = 2.1;
    }
    for (int i = 0; i < SIZE; i++) {
        p_pinned_a[i * SIZE] = 3.2;
        p_pinned_a[i * SIZE + (SIZE -1)] = 4.1;
    }

	clFlush(main_command_queue);
	clFinish(main_command_queue);

	exec_clock = omp_get_wtime();

	clReleaseEvent(event_write);
	clEnqueueWriteBuffer(write_command_queue, mem_a, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_a, 0, NULL, &event_write);
		
    clReleaseEvent(event_write);
	clEnqueueWriteBuffer(write_command_queue, mem_b, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_b, 0, NULL, &event_write);

   	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);
	
	exec_clock = omp_get_wtime();

	for (i = 1; i <= N_ITER; i++) {
		if ((i % 2) == 0) {
			clSetKernelArg(kernel, 0, sizeof(cl_int), &SIZE);
			clSetKernelArg(kernel, 1, sizeof(cl_mem), &mem_a);
			clSetKernelArg(kernel, 2, sizeof(cl_mem), &mem_b);
		} else {
			clSetKernelArg(kernel, 0, sizeof(cl_int), &SIZE);
			clSetKernelArg(kernel, 1, sizeof(cl_mem), &mem_b);
			clSetKernelArg(kernel, 2, sizeof(cl_mem), &mem_a);			
		}

		clReleaseEvent(event_kernel);
		if((i % 10) == 2) {
			clEnqueueNDRangeKernel(main_command_queue, kernel, 2, NULL, global_size, local_size, 1, &event_read, &event_kernel);
		} else {
			clEnqueueNDRangeKernel(main_command_queue, kernel, 2, NULL, global_size, local_size, 0, NULL, &event_kernel);
		}
		clFlush(main_command_queue);

		if ((i % 10) == 0) {
			clReleaseEvent(event_read);
			clEnqueueReadBuffer(read_command_queue, mem_a, CL_FALSE, 0, MATRIX_SIZE, (void *)p_pinned_a, 1, &event_kernel, &event_read);
			clFlush(read_command_queue);
		}
	}

	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clFlush(read_command_queue);
	clFinish(read_command_queue);

	clFlush(write_command_queue);
	clFinish(write_command_queue);
	
	exec_clock = omp_get_wtime() - exec_clock ;

	sum = result = 0;
	for (i = 0; i < SIZE; i++) {
		for (j = 0; j < SIZE; j++) {
			sum += pow(p_pinned_a[i * SIZE + j], 2);
		}
	}
	result = sqrt(sum);

	printf("\n ---------------------- RESULT ---------------------- \n");
	printf("\n Sum: %lf", sum);
	printf("\n Result: %lf", result);
	printf("\n\n ---------------------------------------------------- \n");

	clReleaseEvent(event_write);
	clReleaseEvent(event_read);
	clReleaseEvent(event_kernel);
	clReleaseEvent(event_base);

	clEnqueueUnmapMemObject(main_command_queue, mem_pinned_a, p_pinned_a, 0, NULL, NULL);
	clEnqueueUnmapMemObject(main_command_queue, mem_pinned_b, p_pinned_b, 0, NULL, NULL);
	
	clFlush(main_command_queue);
	clFinish(main_command_queue);

	clReleaseMemObject(mem_pinned_a);
	clReleaseMemObject(mem_a);

	clReleaseMemObject(mem_pinned_b);
	clReleaseMemObject(mem_b);

	clReleaseKernel(kernel);
	clReleaseProgram(program);
	
	clReleaseCommandQueue(main_command_queue);
	clReleaseCommandQueue(read_command_queue);
	clReleaseCommandQueue(write_command_queue);
	
	clReleaseContext(context);

	main_clock = omp_get_wtime() - main_clock ;
	
	printf("\n ---------------------- TIMERS ---------------------- \n");
	printf(" Clock main: %lf\n", main_clock);
	printf(" Clock exec: %lf\n", exec_clock);
	printf("\n\n ---------------------------------------------------- \n");
	
	return 0;
}

