#define CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <CL/cl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h> //For exit()
#include <math.h>
#include <omp.h>

#define SEED 6834723

#ifndef LOCAL_SIZE_0
#define LOCAL_SIZE_0 8
#endif

#ifndef LOCAL_SIZE_1
#define LOCAL_SIZE_1 8
#endif

#ifdef _CTRL_EXAMPLES_OPENCL_GPU_ERROR_CHECK_
	#include <assert.h>
	
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_DEBUG_
		#define OPENCL_ASSERT_OP( operation ) \
			err = operation; \
			printf("error: %d\n", err); \
			assert(err == CL_SUCCESS);

		#define OPENCL_ASSERT_ERROR( err ) \
			printf("error: %d\n", err); \
			assert(err == CL_SUCCESS)

	#else 
		#define OPENCL_ASSERT_OP( operation ) \
			assert(operation == CL_SUCCESS);

		#define OPENCL_ASSERT_ERROR( err ) \
			assert(err == CL_SUCCESS)
	#endif
#else
	#define OPENCL_ASSERT_OP( operation ) operation
	#define OPENCL_ASSERT_ERROR(err)
#endif



#define OPENCL_PROFILE_OP_READ 1
#define OPENCL_PROFILE_OP_WRITE 2
#define OPENCL_PROFILE_OP_UPDATE 3
#define OPENCL_PROFILE_OP_COPY 4

#define OPENCL_PROFILE_QUEUE_MAIN 0
#define OPENCL_PROFILE_QUEUE_READ 1
#define OPENCL_PROFILE_QUEUE_WRITE 2

#ifdef _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
	#define OPENCL_PROFILE_READ( event )	\
		profiling_read_events[i_read_task] = event; \
		OPENCL_ASSERT_OP( clRetainEvent(profiling_read_events[i_read_task]) ); \
		i_read_task++;

	#define OPENCL_PROFILE_WRITE( event )	\
		profiling_write_events[i_write_task] = event; \
		OPENCL_ASSERT_OP( clRetainEvent(profiling_write_events[i_write_task]) ); \
		i_write_task++;

	#define OPENCL_PROFILE_UPDATE( event ) \
		profiling_update_events[i_update_task] = event; \
		OPENCL_ASSERT_OP( clRetainEvent(profiling_update_events[i_update_task]) ); \
		i_update_task++;

	#define OPENCL_PROFILE_COPY( event ) \
		profiling_copy_events[i_copy_task] = event; \
		OPENCL_ASSERT_OP( clRetainEvent(profiling_copy_events[i_copy_task]) ); \
		i_copy_task++;

	#define OPENCL_PROFILE_FIRST( event ) \
		first_profiling_event = event; \
		OPENCL_ASSERT_OP( clRetainEvent(first_profiling_event) );

	#define OPENCL_PROFILE_LAST( event ) \
		last_profiling_event = event; \
		OPENCL_ASSERT_OP( clRetainEvent(last_profiling_event) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_GENERATE_VISUAL_PROFILER_
		typedef struct {
			cl_event event;
			int queue;
			int	op;
		} visual_event;

		#define OPENCL_PROFILE_VISUAL( command_queue, visual_event, operation ) \
			profiling_visual_events[i_visual_task].event = visual_event; \
			profiling_visual_events[i_visual_task].queue = command_queue; \
			profiling_visual_events[i_visual_task].op = operation; \
			OPENCL_ASSERT_OP( clRetainEvent(profiling_visual_events[i_visual_task].event) ); \
			i_visual_task++;	
	#else 
		#define OPENCL_PROFILE_VISUAL( queue, event, op ) 
	#endif
#else
	#define OPENCL_PROFILE_READ( event )
	#define OPENCL_PROFILE_WRITE( event )
	#define OPENCL_PROFILE_UPDATE( event )
	#define OPENCL_PROFILE_COPY( event )
	#define OPENCL_PROFILE_FIRST( event )
	#define OPENCL_PROFILE_LAST( event )
	#define OPENCL_PROFILE_VISUAL( queue, event, op )
#endif

double main_clock;
double exec_clock;

#define CELL_AUTOM_KERNEL_NAME_UPDATE "Update"

#define CELL_AUTOM_KERNEL_UPDATE \
    "__kernel void Update(const int SIZE, __global float *d_update, __global const float *d_copy) { \
        int row = get_global_id(1) + 1; \
		int col = get_global_id(0) + 1; \
        \
         if (!(row >= SIZE - 1 || col >= SIZE - 1)) { \
			d_update[row * SIZE + col] = \
			( \
				d_copy[(row - 1) * SIZE + col] + \
				d_copy[(row + 1) * SIZE + col] + \
				d_copy[row * SIZE + (col - 1)] + \
				d_copy[row * SIZE + (col + 1)] \
			) / 4.0; \
		} \
    } \n\n"


#define CELL_AUTOM_KERNEL_NAME_COPY "Copy"

#define CELL_AUTOM_KERNEL_COPY \
    "__kernel void Copy(const int SIZE, __global float *d_copy, __global const float *d_update) { \
		int row = get_global_id(1); \
		int col = get_global_id(0); \
        \
		if (row < SIZE && col < SIZE) { \
            d_copy[row * SIZE + col] = d_update[row * SIZE + col]; \
        } \
    } \n\n"


int main (int argc, char *argv[]) {
	main_clock = omp_get_wtime();

	/* ARGUMENTS */

	if (argc < 5) {
		fprintf(stderr, "Usage: %s <size> <numIter> <platform> <device>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	cl_int SIZE = atoi(argv[1]);
	cl_long N_ITER = atoi(argv[2]);
	cl_int PLATFORM = atoi(argv[3]);
	cl_int DEVICE = atoi(argv[4]);

	size_t MATRIX_SIZE = sizeof(cl_float) * SIZE * SIZE;
	size_t MATRIX_TAM = SIZE * SIZE;

	cl_int n_read_taks = N_ITER;
	cl_int n_write_taks = 1;
	cl_int n_update_taks = N_ITER;
	cl_int n_copy_taks = N_ITER;

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
			printf("%ld %ld %d %d\n", SIZE, N_ITER, PLATFORM, DEVICE);
		#else
			printf("\n ----------------------- ARGS ----------------------- \n");
			printf("\n SIZE: %ld", SIZE);
			printf("\n N_ITER: %ld", N_ITER);
			printf("\n PLATFORM: %d", PLATFORM);
			printf("\n DEVICE: %d", DEVICE);
			printf("\n TOTAL TASKS: %d", n_update_taks + n_copy_taks + n_read_taks + n_write_taks);
			printf("\n KERNEL TASKS: %d", n_update_taks + n_copy_taks);
			printf("\n UPDATE TASKS: %d", n_update_taks);
			printf("\n COPY TASKS: %d", n_copy_taks);
			printf("\n OFFLOADING TASKS: %d", n_read_taks + n_write_taks);
			printf("\n READ TASKS: %d", n_read_taks);
			printf("\n WRITE TASKS: %d", n_write_taks);
			#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
				printf("\n POLICY ASYNC");
			#else // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
				printf("\n POLICY SYNC");
			#endif // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			printf("\n\n ---------------------------------------------------- \n");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

	/* VARIABLES */

	int i, j, k;
	cl_int err;

    char *kernel_raw_update = CELL_AUTOM_KERNEL_UPDATE;
	char *kernel_raw_copy = CELL_AUTOM_KERNEL_COPY;

	char *kernel_raw = (char *)malloc((strlen(kernel_raw_update) + strlen(kernel_raw_copy)) * sizeof(char));
	sprintf(&kernel_raw[0], "%s", kernel_raw_update);
	sprintf(&kernel_raw[strlen(kernel_raw)], "%s", kernel_raw_copy);
	// strcat(kernel_raw, kernel_raw_update);
	// strcat(kernel_raw, kernel_raw_copy);

	size_t kernel_size = strlen(kernel_raw);

	size_t local_size[2];
	size_t global_size[2];

	double result;
	double sum;

	cl_platform_id platform_id;
	cl_device_id device_id;

	cl_context context;
	cl_program program;
	cl_kernel kernel_update;
	cl_kernel kernel_copy;

	cl_command_queue_properties properties;
	
	cl_command_queue main_command_queue;
	cl_command_queue write_command_queue;
	cl_command_queue read_command_queue;

	cl_mem mem_update;
	cl_mem mem_copy;

	cl_mem pinned_update;

	cl_float *p_pinned_update;

	cl_event event_read;
	cl_event event_write;
	cl_event event_copy;
	cl_event event_update;

	cl_event event_base;

	

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
		cl_profiling_info profiling_info_start = CL_PROFILING_COMMAND_START;
		cl_profiling_info profiling_info_end = CL_PROFILING_COMMAND_END;

		cl_event first_profiling_event;
		cl_event last_profiling_event;

		cl_event *profiling_read_events = (cl_event *)malloc(n_read_taks * sizeof(cl_event));
		cl_event *profiling_write_events = (cl_event *)malloc(n_write_taks * sizeof(cl_event));
		cl_event *profiling_update_events = (cl_event *)malloc(n_update_taks * sizeof(cl_event));
		cl_event *profiling_copy_events = (cl_event *)malloc(n_copy_taks * sizeof(cl_event));

		int i_read_task = 0;
		int i_write_task = 0;
		int i_update_task = 0;
		int i_copy_task = 0;


		cl_ulong profiling_total = 0;
		cl_ulong profiling_sum = 0;
		cl_ulong profiling_offloading = 0;
		cl_ulong profiling_read = 0;
		cl_ulong profiling_write = 0;
		cl_ulong profiling_kernel = 0;
		cl_ulong profiling_update = 0;
		cl_ulong profiling_copy = 0;

		cl_ulong profiling_start = 0;
		cl_ulong profiling_end = 0;
		
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_GENERATE_VISUAL_PROFILER_
			int n_visual_task = n_update_taks + n_copy_taks + n_read_taks + n_write_taks;
			int i_visual_task = 0;

			visual_event *profiling_visual_events = (visual_event *)malloc(n_visual_task * sizeof(visual_event));
		#endif
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_

	/* PLATFORMS & DEVICES */

    cl_platform_id *p_platforms = (cl_platform_id *)malloc((PLATFORM + 1) * sizeof(cl_platform_id));
	OPENCL_ASSERT_OP( clGetPlatformIDs(PLATFORM + 1, p_platforms, NULL) );
    platform_id = p_platforms[PLATFORM];
	free(p_platforms);

    cl_device_id *p_devices = (cl_device_id *)malloc((DEVICE + 1) * sizeof(cl_device_id));
    OPENCL_ASSERT_OP( clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, DEVICE + 1, p_devices, NULL) );
    device_id = p_devices[DEVICE];
    free(p_devices);
	
	/* SET UP, CONTEXTO, COLAS, KERNELS, ETC */

	cl_context_properties context_properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_id, 0 };
	context = clCreateContext(context_properties, 1, &device_id, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	properties = 0;

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
		properties |= CL_QUEUE_PROFILING_ENABLE;
	#endif

	main_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
	OPENCL_ASSERT_ERROR( err );
	
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		read_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
		OPENCL_ASSERT_ERROR( err );
		write_command_queue = clCreateCommandQueue(context, device_id, properties, &err);
		OPENCL_ASSERT_ERROR( err );
	#endif

	pinned_update = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	p_pinned_update = (float *)clEnqueueMapBuffer(main_command_queue, pinned_update, CL_FALSE, CL_MAP_READ | CL_MAP_WRITE, 0, MATRIX_SIZE, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	

	mem_update = clCreateBuffer(context, CL_MEM_READ_WRITE, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	mem_copy = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_HOST_NO_ACCESS, MATRIX_SIZE, NULL, &err);
	OPENCL_ASSERT_ERROR( err );


	program = clCreateProgramWithSource(context, 1, (const char **)(&kernel_raw), (const size_t *)(&kernel_size), &err);
	OPENCL_ASSERT_ERROR( err );
	
	OPENCL_ASSERT_OP( clBuildProgram(program, 1, &device_id, NULL, NULL, NULL) );
	// err = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
	// if (err == CL_BUILD_PROGRAM_FAILURE) {
	// 	size_t log_size;
	// 	clGetProgramBuildInfo(program, device_id,CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
	// 	char *log = (char *) malloc(log_size);
	// 	clGetProgramBuildInfo(program, device_id,CL_PROGRAM_BUILD_LOG, log_size, log, NULL);
	// 	printf("%s\n", log); fflush(stdout);
	// 	free(log);
	// }
	
	kernel_update = clCreateKernel(program, CELL_AUTOM_KERNEL_NAME_UPDATE, &err);
	OPENCL_ASSERT_ERROR( err );
	
	kernel_copy = clCreateKernel(program, CELL_AUTOM_KERNEL_NAME_COPY, &err);
	OPENCL_ASSERT_ERROR( err );

	err = clSetKernelArg(kernel_copy, 0, sizeof(cl_int), &SIZE);
	err |= clSetKernelArg(kernel_copy, 1, sizeof(cl_mem), &mem_copy);
	err |= clSetKernelArg(kernel_copy, 2, sizeof(cl_mem), &mem_update);
	OPENCL_ASSERT_ERROR( err );

	err = clSetKernelArg(kernel_update, 0, sizeof(cl_int), &SIZE);
	err |= clSetKernelArg(kernel_update, 1, sizeof(cl_mem), &mem_update);
	err |= clSetKernelArg(kernel_update, 2, sizeof(cl_mem), &mem_copy);
	OPENCL_ASSERT_ERROR( err );
	

	local_size[0] = LOCAL_SIZE_0;
	local_size[1] = LOCAL_SIZE_1;

	global_size[0] = global_size[1] = SIZE;

	if ((SIZE % local_size[0]) != 0) {
		global_size[0] += (local_size[0] - (SIZE % local_size[0]));
	}

	if ((SIZE % local_size[1]) != 0) {
		global_size[1] += (local_size[1] - (SIZE % local_size[1]));
	}

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
			printf("%lu %lu %lu %lu\n", local_size[0], local_size[1], global_size[0], global_size[1]);
		#else
			printf("\n ------------------ WORK-GROUP-INFO ----------------- \n");
			printf("\n Local Size: %lu x %lu", local_size[0], local_size[1]);
			printf("\n Global Size: %lu x %lu", global_size[0], global_size[1]);
			printf("\n\n ---------------------------------------------------- \n");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

	
	event_base = clCreateUserEvent(context, &err);
	OPENCL_ASSERT_ERROR( err );

	OPENCL_ASSERT_OP( clSetUserEventStatus(event_base, CL_COMPLETE) );

	event_copy = event_base;
	event_read = event_base;
	event_update = event_base;

	OPENCL_ASSERT_OP( clRetainEvent(event_update) );
	OPENCL_ASSERT_OP( clRetainEvent(event_copy) );
	OPENCL_ASSERT_OP( clRetainEvent(event_read) );

	srand(SEED);

	for (i = 0; i < SIZE; i++) {
		for (j = 0; j < SIZE; j++) {
			p_pinned_update[i * SIZE + j] = 0.0;
		}
	}

	for (j = 0; j < SIZE; j++) {
		p_pinned_update[j] = 1.0;
		p_pinned_update[(SIZE - 1) * SIZE + j] = 2.0;
	}
	for (i = 0; i < SIZE; i++) {
		p_pinned_update[i * SIZE] = 3.0;
		p_pinned_update[i * SIZE + SIZE - 1] = 4.0;
	}

	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(write_command_queue, mem_update, CL_FALSE, 0, MATRIX_SIZE,(void *)&p_pinned_update[0], 0, NULL, &event_write) );
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_WRITE, event_write, OPENCL_PROFILE_OP_WRITE );
	#else
		OPENCL_ASSERT_OP( clEnqueueWriteBuffer(main_command_queue, mem_update, CL_FALSE, 0, MATRIX_SIZE,(void *)&p_pinned_update[0], 0, NULL, &event_write) );
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_write, OPENCL_PROFILE_OP_WRITE );
	#endif
	OPENCL_PROFILE_WRITE( event_write );
	OPENCL_PROFILE_FIRST( event_write );
	
	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif

	exec_clock = omp_get_wtime();

	for (i = 0; i < N_ITER; i++) {
			
		OPENCL_ASSERT_OP( clReleaseEvent(event_copy) );
		if (i != 0) {
			OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_copy, 2, NULL, global_size, local_size, 1, &event_update, &event_copy) );
		} else {
			OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_copy, 2, NULL, global_size, local_size, 1, &event_write, &event_copy) );
		}
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_copy, OPENCL_PROFILE_OP_COPY );
		OPENCL_PROFILE_COPY( event_copy );


		OPENCL_ASSERT_OP( clReleaseEvent(event_update) );
		OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(main_command_queue, kernel_update, 2, NULL, global_size, local_size, 1, &event_copy, &event_update) );
		OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_update, OPENCL_PROFILE_OP_UPDATE );
		OPENCL_PROFILE_UPDATE( event_update );

		OPENCL_ASSERT_OP( clReleaseEvent(event_read) );

		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			OPENCL_ASSERT_OP( clFlush(main_command_queue) );
		#endif
		
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			OPENCL_ASSERT_OP( clEnqueueReadBuffer(read_command_queue, mem_update, CL_FALSE, 0, MATRIX_SIZE, (void *)&p_pinned_update[0], 1, &event_update, &event_read) );
			OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_READ, event_read, OPENCL_PROFILE_OP_READ );
		#else
			OPENCL_ASSERT_OP( clEnqueueReadBuffer(main_command_queue, mem_update, CL_FALSE, 0, MATRIX_SIZE, (void *)&p_pinned_update[0], 1, &event_update, &event_read) );
			OPENCL_PROFILE_VISUAL( OPENCL_PROFILE_QUEUE_MAIN, event_read, OPENCL_PROFILE_OP_READ );
		#endif
		OPENCL_PROFILE_READ( event_read );

		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		#endif
	}

	OPENCL_PROFILE_LAST( event_read );

	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clFlush(read_command_queue) );
		OPENCL_ASSERT_OP( clFinish(read_command_queue) );

		OPENCL_ASSERT_OP( clFlush(write_command_queue) );
		OPENCL_ASSERT_OP( clFinish(write_command_queue) );
	#endif

	exec_clock = omp_get_wtime() - exec_clock ;

	/* CALCULO RESULTADOS */

	sum = result = 0;
	for (i = 0; i < SIZE; i++) {
		for (j = 0; j < SIZE; j++) {
			sum += pow(p_pinned_update[i * SIZE + j], 2);
		}
	}
	result = sqrt(sum);

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
		printf("%lf %lf\n", sum, result);
	#else 
		printf("\n ---------------------- RESULT ---------------------- \n");
		printf("\n Sum: %lf", sum);
		printf("\n Result: %lf", result);
		printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_

	
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_GENERATE_VISUAL_PROFILER_
			FILE *f = fopen("visual_profiler_info.txt", "w");

			for (i = 0; i < n_visual_task; i++) {
				OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_visual_events[i].event, profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
				OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_visual_events[i].event, profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );

				fprintf(f, "%d; %d; %lu; %lu\n", profiling_visual_events[i].queue, profiling_visual_events[i].op, profiling_start, profiling_end);
				OPENCL_ASSERT_OP( clReleaseEvent(profiling_visual_events[i].event) );
			}

			fclose(f);
			free(profiling_visual_events);
		#endif
		for (i = 0; i < n_read_taks; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_read_events[i], profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_read_events[i], profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );

			profiling_sum += profiling_end - profiling_start;
			profiling_read += profiling_end - profiling_start;
			profiling_offloading += profiling_end - profiling_start;
			
			OPENCL_ASSERT_OP( clReleaseEvent(profiling_read_events[i]) );
		}
		free(profiling_read_events);

		for (i = 0; i < n_write_taks; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_write_events[i], profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_write_events[i], profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );

			profiling_sum += profiling_end - profiling_start;
			profiling_write += profiling_end - profiling_start;
			profiling_offloading += profiling_end - profiling_start;
			
			OPENCL_ASSERT_OP( clReleaseEvent(profiling_write_events[i]) );
		}
		free(profiling_write_events);

		for (i = 0; i < n_update_taks; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_update_events[i], profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_update_events[i], profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );
			
			profiling_sum += profiling_end - profiling_start;
			profiling_update += profiling_end - profiling_start;
			profiling_kernel += profiling_end - profiling_start;

			OPENCL_ASSERT_OP( clReleaseEvent(profiling_update_events[i]) );
		}
		free(profiling_update_events);

		for (i = 0; i < n_copy_taks; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_copy_events[i], profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(profiling_copy_events[i], profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );
			
			profiling_sum += profiling_end - profiling_start;
			profiling_copy += profiling_end - profiling_start;
			profiling_kernel += profiling_end - profiling_start;

			OPENCL_ASSERT_OP( clReleaseEvent(profiling_copy_events[i]) );
		}
		free(profiling_copy_events);

		OPENCL_ASSERT_OP( clGetEventProfilingInfo(first_profiling_event, profiling_info_start, sizeof(cl_ulong), (void *)(&profiling_start), NULL) );
		OPENCL_ASSERT_OP( clGetEventProfilingInfo(last_profiling_event, profiling_info_end, sizeof(cl_ulong), (void *)(&profiling_end), NULL) );

		profiling_total = profiling_end - profiling_start;

		OPENCL_ASSERT_OP( clReleaseEvent(first_profiling_event) );
		OPENCL_ASSERT_OP( clReleaseEvent(last_profiling_event) );

		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
			printf("%.3lf %.3lf %.3lf %.3lf %.3lf\n", profiling_total, profiling_sum, profiling_kernel, profiling_offloading, profiling_read, profiling_write);
		#else
			printf("\n -------------------- PROFILING --------------------- \n");
			printf("\n Total Time: %.3lf", profiling_total / 1000000.0);
			printf("\n Acumulate Sum Time: %.3lf", profiling_sum / 1000000.0);
			printf("\n Kernel Time: %.3lf", profiling_kernel / 1000000.0);
			printf("\n Update Time: %.3lf", profiling_update / 1000000.0);
			printf("\n Copy Time: %.3lf", profiling_copy / 1000000.0);
			printf("\n Offloading Time: %.3lf", profiling_offloading / 1000000.0);
			printf("\n Read Time: %.3lf", profiling_read / 1000000.0);
			printf("\n Write Time: %.3lf", profiling_write / 1000000.0);
			printf("\n\n ---------------------------------------------------- \n");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_

	/* RELEASE ZONE */

	
	OPENCL_ASSERT_OP( clReleaseEvent(event_copy) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_update) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_read) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_write) );
	OPENCL_ASSERT_OP( clReleaseEvent(event_base) );

	OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(main_command_queue, pinned_update, p_pinned_update, 0, NULL, NULL) );
	
	OPENCL_ASSERT_OP( clFlush(main_command_queue) );
	OPENCL_ASSERT_OP( clFinish(main_command_queue) );

	OPENCL_ASSERT_OP( clReleaseMemObject(pinned_update) );
	OPENCL_ASSERT_OP( clReleaseMemObject(mem_update) );
	OPENCL_ASSERT_OP( clReleaseMemObject(mem_copy) );

	OPENCL_ASSERT_OP( clReleaseKernel(kernel_update) );
	OPENCL_ASSERT_OP( clReleaseKernel(kernel_copy) );
	OPENCL_ASSERT_OP( clReleaseProgram(program) );
	
	OPENCL_ASSERT_OP( clReleaseCommandQueue(main_command_queue) );
	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		OPENCL_ASSERT_OP( clReleaseCommandQueue(read_command_queue) );
		OPENCL_ASSERT_OP( clReleaseCommandQueue(write_command_queue) );
	#endif
	
	OPENCL_ASSERT_OP( clReleaseContext(context) );

	main_clock = omp_get_wtime() - main_clock ;

	#ifdef _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
		printf("%lf %lf\n", main_clock, exec_clock);
	#else 
		printf("\n ---------------------- TIMERS ---------------------- \n");
		printf("Clock main: %lf\n", main_clock);
		printf("Clock exec: %lf\n", exec_clock);
		printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	
	return 0;
}

